from collections.abc import Callable
import sys
from typing import Any

from esphome import codegen as cg, config_validation as cv
from esphome.automation import register_action
from esphome.config_validation import Invalid, Schema
from esphome.const import (
    CONF_DEFAULT,
    CONF_GROUP,
    CONF_ID,
    CONF_MAX_VALUE,
    CONF_MIN_VALUE,
    CONF_STATE,
    CONF_TYPE,
)
from esphome.core import ID, EsphomeError, TimePeriod
from esphome.coroutine import FakeAwaitable
from esphome.cpp_generator import MockObj
from esphome.schema_extractors import EnableSchemaExtraction
from esphome.types import Expression

from ..defines import (
    CONF_FLEX_ALIGN_CROSS,
    CONF_FLEX_ALIGN_MAIN,
    CONF_FLEX_ALIGN_TRACK,
    CONF_FLEX_FLOW,
    CONF_GRID_COLUMN_ALIGN,
    CONF_GRID_COLUMNS,
    CONF_GRID_ROW_ALIGN,
    CONF_GRID_ROWS,
    CONF_LAYOUT,
    CONF_MAIN,
    CONF_PAD_COLUMN,
    CONF_PAD_ROW,
    CONF_SCALE,
    CONF_STYLES,
    CONF_WIDGETS,
    LOGGER,
    OBJ_FLAGS,
    PARTS,
    STATES,
    LValidator,
    add_lv_use,
    call_lambda,
    get_styles_used,
    get_theme_widget_map,
    get_widget_map,
    get_widgets_completed,
    join_enums,
    literal,
)
from ..lv_validation import lv_int
from ..lvcode import (
    LvConditional,
    add_line_marks,
    lv,
    lv_add,
    lv_assign,
    lv_expr,
    lv_obj,
    lv_Pvariable,
    lvgl_static,
)
from ..types import (
    LV_STATE,
    LvCompound,
    LvType,
    ObjUpdateAction,
    lv_coord_t,
    lv_obj_t,
    lv_obj_t_ptr,
)

EVENT_LAMB = "event_lamb__"


def _build_update_schema(widget_type: "WidgetType") -> Schema:
    # Local import: ..schemas imports WidgetType from this module.
    from ..schemas import base_update_schema

    return base_update_schema(widget_type, widget_type.parts).extend(
        widget_type.modify_schema
    )


def _update_action_schema(
    widget_type: "WidgetType",
) -> Schema | Callable[[Any], Any]:
    # Eager when extracting so build_language_schema.py sees the mapping;
    # lazy otherwise to skip ~200 ms of import-time voluptuous work.
    if EnableSchemaExtraction:
        return _build_update_schema(widget_type)

    cached: Schema | None = None

    def validator(value: Any) -> Any:
        nonlocal cached
        if cached is None:
            cached = _build_update_schema(widget_type)
        return cached(value)

    return validator


class WidgetType:
    """
    Describes a type of Widget, e.g. "bar" or "line"
    """

    def __init__(
        self,
        name: str,
        w_type: LvType,
        parts: tuple,
        schema=None,
        modify_schema=None,
        lv_name=None,
        is_mock: bool = False,
    ):
        """
        :param name: The widget name, e.g. "bar"
        :param w_type: The C type of the widget
        :param parts: What parts this widget supports
        :param schema: The config schema for defining a widget
        :param modify_schema: A schema to update the widget, defaults to the same as the schema
        :param lv_name: The name of the LVGL widget in the LVGL library, if different from the name
        :param is_mock: Whether this widget is a mock widget, i.e. not a real LVGL widget
        """
        self.name = name
        self.lv_name = lv_name or name
        self.w_type = w_type
        self.parts = parts
        if not isinstance(schema, Schema):
            schema = Schema(schema or {})
        self.schema = schema
        if modify_schema is None:
            modify_schema = schema
        if not isinstance(modify_schema, Schema):
            modify_schema = Schema(modify_schema)
        self.modify_schema = modify_schema
        self.mock_obj = MockObj(f"lv_{self.lv_name}", "_")

        # Local import to avoid circular import
        from ..automation import update_to_code
        from ..schemas import WIDGET_TYPES

        if not is_mock:
            if self.name in WIDGET_TYPES:
                raise EsphomeError(f"Duplicate definition of widget type '{self.name}'")
            WIDGET_TYPES[self.name] = self

            register_action(
                f"lvgl.{self.name}.update",
                ObjUpdateAction,
                _update_action_schema(self),
                synchronous=True,
            )(update_to_code)

    @property
    def animated(self):
        return False

    @property
    def required_component(self):
        return None

    def is_compound(self):
        return self.w_type.inherits_from(LvCompound)

    async def create_to_code(self, config: dict, parent: MockObj) -> "Widget":
        """
        Generate code for a widget creation.
        :param config: The configuration for the widget
        :param parent: The parent to which it should be attached
        """

        creator = await self.obj_creator(parent, config)
        add_lv_use(self.name)
        add_lv_use(*self.get_uses())
        wid = config[CONF_ID]
        add_line_marks(wid)
        if self.is_compound():
            var = cg.new_Pvariable(wid)
            lv_add(var.set_obj(creator))
            await self.on_create(var.obj, config)
        else:
            var = lv_Pvariable(lv_obj_t, wid)
            lv_assign(var, creator)
            await self.on_create(var, config)

        w = Widget.create(wid, var, self, config)
        if theme := get_theme_widget_map().get(self.name):
            for part, states in theme.items():
                part = "LV_PART_" + part.upper()
                for state, style in states.items():
                    state = "LV_STATE_" + state.upper()
                    if state == "LV_STATE_DEFAULT":
                        lv_state = literal(part)
                    elif part == "LV_PART_MAIN":
                        lv_state = literal(state)
                    else:
                        lv_state = join_enums((state, part))
                    w.add_style(style, lv_state)
        await set_obj_properties(w, config)
        await add_widgets(w, config)
        await self.to_code(w, config)
        return w

    async def to_code(self, w: "Widget", config: dict):
        """
        Update a widget, also called when creating
        :param config:
        :return:
        """

    async def obj_creator(self, parent: MockObj, config: dict):
        """
        Create an instance of the widget type
        :param parent: The parent to which it should be attached
        :param config:  Its configuration
        :return: Generated code as a single text line
        """
        return lv_expr.call(f"{self.lv_name}_create", parent)

    async def on_create(self, var: MockObj, config: dict):
        """
        Called from to_code when the widget is created, to set up any initial properties
        :param var: The variable representing the widget
        :param config: Its configuration
        """

    def get_uses(self):
        """
        Get a list of other widgets used by this one
        :return:
        """
        return ()

    async def get_max(self, config: dict):
        return sys.maxsize

    async def get_min(self, config: dict):
        return -sys.maxsize

    def get_step(self, config: dict):
        return 1

    def get_scale(self, config: dict):
        return 1.0

    def validate(self, value):
        """
        Provides an opportunity for custom validation for a given widget type
        :param value:
        :return:
        """
        return value

    def final_validate(self, widget, update_config, widget_config, path):
        """
        Allow final validation for a given widget type update action
        :param widget: A widget
        :param update_config: The configuration for the update action
        :param widget_config: The configuration for the widget itself
        :param path: The path to the widget, for error reporting
        """


class Widget:
    """
    Represents a Widget.
    This class has a lot of methods. Adding any more runs foul of lint checks ("too many public methods").
    """

    def __init__(self, var, wtype: WidgetType, config: dict = None):
        self.var = var
        self.type = wtype
        self.config = config
        self.scale = 1.0
        self.range_from = -sys.maxsize
        self.range_to = sys.maxsize
        if wtype.is_compound():
            self.obj = MockObj(f"{self.var}->obj")
        else:
            self.obj = var
        self.outer = None
        self.move_to_foreground = False
        # Properties for linear equations
        self.slope = None
        self.y_int = None
        self.parent = None

    @staticmethod
    def create(name, var, wtype: WidgetType, config: dict = None):
        w = Widget(var, wtype, config)
        get_widget_map()[name] = w
        return w

    def set_state(self, state: MockObj, value: bool | Expression):
        lv_add(lvgl_static.lv_obj_set_state_value(self.obj, state, value))

    def has_state(self, state: MockObj):
        return lv_expr.obj_has_state(self.obj, state)

    def is_pressed(self):
        return self.has_state(LV_STATE.PRESSED)

    def is_checked(self):
        return self.has_state(LV_STATE.CHECKED)

    def add_flag(self, flag):
        if "|" in flag:
            flag = f"(lv_obj_flag_t)({flag})"
        return lv_obj.add_flag(self.obj, literal(flag))

    def clear_flag(self, flag):
        if "|" in flag:
            flag = f"(lv_obj_flag_t)({flag})"
        return lv_obj.remove_flag(self.obj, literal(flag))

    def add_style(self, style_id, state=LV_STATE.DEFAULT):
        if "|" in state:
            state = f"(lv_state_t)({state})"
        lv_obj.add_style(self.obj, MockObj(style_id), literal(state))

    async def set_property(
        self, prop, value, animated: bool = None, lv_name=None, processor=None
    ):
        """
        Set a property of the widget.
        :param prop:  The property name
        :param value:  The value
        :param animated:  If the change should be animated
        :param lv_name:  The base type of the widget e.g. "obj"
        """

        from ..schemas import ALL_STYLES, remap_property

        if isinstance(value, dict):
            value = value.get(prop)
            if value is None:
                return
            if not processor and isinstance(ALL_STYLES.get(prop), LValidator):
                processor = ALL_STYLES[prop]
            if isinstance(processor, LValidator):
                processor = processor.process
            if processor:
                value = await processor(value)
        elif value is None:
            return
        prop = remap_property(prop)
        if isinstance(value, TimePeriod):
            value = value.total_milliseconds
        elif isinstance(value, str):
            value = literal(value)
        elif isinstance(value, ID):
            value = MockObj(value)
        lv_name = lv_name or self.type.lv_name
        if animated is None or self.type.animated is not True:
            lv.call(f"{lv_name}_set_{prop}", self.obj, value)
        else:
            lv.call(
                f"{lv_name}_set_{prop}",
                self.obj,
                value,
                literal("LV_ANIM_ON" if animated else "LV_ANIM_OFF"),
            )

    def get_property(self, prop, ltype=None):
        ltype = ltype or self.__type_base()
        return cg.RawExpression(f"lv_{ltype}_get_{prop}({self.obj})")

    def set_style(self, prop: str, value, state=LV_STATE.DEFAULT):
        if value is None:
            return
        get_styles_used().add(prop)
        if isinstance(value, str):
            value = literal(value)
        lv.call(f"obj_set_style_{prop}", self.obj, value, state)

    def __type_base(self):
        wtype = self.type.w_type
        base = str(wtype)
        if base.startswith("Lv"):
            return f"{wtype}".removeprefix("Lv").removesuffix("Type").lower()
        return f"{wtype}".removeprefix("lv_").removesuffix("_t")

    def __str__(self):
        return f"({self.var}, {self.type})"

    def get_args(self):
        if isinstance(self.type.w_type, LvType):
            return self.type.w_type.args.copy()
        return [(lv_obj_t_ptr, "obj")]

    def get_value(self):
        if isinstance(self.type.w_type, LvType):
            result = self.type.w_type.value(self)
            if isinstance(result, list):
                return result[0]
            return result
        return self.obj

    def get_values(self):
        if isinstance(self.type.w_type, LvType):
            result = self.type.w_type.value(self)
            if isinstance(result, list):
                return result
            return [result]
        return [self.obj]

    def get_number_value(self):
        value = self.type.mock_obj.get_value(self.obj)
        if self.scale == 1.0:
            return value
        return value / float(self.scale)

    def is_selected(self):
        """
        Overridable property to determine if the widget is selected. Will be None except
        for matrix buttons
        :return:
        """
        return

    def get_scale(self):
        return self.type.get_scale(self.config)


class LvScrActType(WidgetType):
    """
    A "widget" representing the active screen.
    """

    def __init__(self):
        super().__init__("lv_screen_active()", lv_obj_t, (), is_mock=True)

    async def to_code(self, w, config: dict):
        pass


def get_screen_active(lv_comp: MockObj) -> Widget:
    return Widget(lv_comp.get_screen_active(), LvScrActType(), {})


def get_widget_generator(wid):
    """
    Used to wait for a widget during code generation.
    :param wid:
    :return:
    """
    widget_map = get_widget_map()
    while True:
        if obj := widget_map.get(wid):
            return obj
        if get_widgets_completed():
            raise Invalid(
                f"Widget {wid} not found, yet all widgets should be defined by now"
            )
        yield


async def get_widget_(wid):
    if obj := get_widget_map().get(wid):
        return obj
    return await FakeAwaitable(get_widget_generator(wid))


def widgets_wait_generator():
    while True:
        if get_widgets_completed():
            return
        yield


async def wait_for_widgets():
    if get_widgets_completed():
        return
    await FakeAwaitable(widgets_wait_generator())


async def get_widgets(config: dict | list, id: str = CONF_ID) -> list[Widget]:
    if not config:
        return []
    if not isinstance(config, list):
        config = [config]
    return [await get_widget_(c[id]) for c in config if id in c]


def collect_props(config):
    """
    Collect all properties from a configuration
    :param config:
    :return:
    """

    from ..schemas import ALL_STYLES

    props = {}
    for prop in [*ALL_STYLES, *OBJ_FLAGS, CONF_STYLES, CONF_GROUP]:
        if prop in config:
            if prop == CONF_SCALE:
                props[CONF_SCALE + "_x"] = config[prop]
                props[CONF_SCALE + "_y"] = config[prop]
            else:
                props[prop] = config[prop]
    return props


def collect_states(config):
    """
    Collect prperties for each state of a widget
    :param config:
    :return:
    """
    states = {CONF_DEFAULT: collect_props(config)}
    for state in STATES:
        if state in config:
            states[state] = collect_props(config[state])
    return states


def collect_parts(config):
    """
    Collect properties and states for all widget parts
    :param config:
    :return:
    """
    parts = {CONF_MAIN: collect_states(config)}
    for part in PARTS:
        if part in config:
            parts[part] = collect_states(config[part])
    return parts


def _size_to_str(value):
    if isinstance(value, float):
        return f"lv_pct({int(value * 100)})"
    return str(value)


def _grid_descriptor_array(name: str, specs) -> MockObj:
    """Generate a file-scope ``static const`` grid row/column descriptor array
    and return a reference to it."""
    values = ",".join(_size_to_str(x) for x in specs)
    initializer = "{" + values + ", LV_GRID_TEMPLATE_LAST}"
    arr_id = ID(name, is_declaration=True, type=lv_coord_t)
    return cg.static_const_array(arr_id, cg.RawExpression(initializer))


def _set_layout_options(w: Widget, layout: dict, base_name: str | None) -> None:
    """Apply the layout options present in ``layout`` to ``w``.

    Only options actually present are applied, so this works both for widget
    creation (where every option is supplied) and for update actions (where the
    layout ``type`` and grid structure are fixed and only the style options are
    changed). ``base_name`` names the generated grid descriptor arrays and is
    only required at creation, when ``grid_rows``/``grid_columns`` are present.
    """
    if (pad_row := layout.get(CONF_PAD_ROW)) is not None:
        w.set_style(CONF_PAD_ROW, pad_row)
    if (pad_column := layout.get(CONF_PAD_COLUMN)) is not None:
        w.set_style(CONF_PAD_COLUMN, pad_column)
    if (rows := layout.get(CONF_GRID_ROWS)) is not None:
        w.set_style(
            "grid_row_dsc_array", _grid_descriptor_array(f"{base_name}_row_dsc", rows)
        )
    if (columns := layout.get(CONF_GRID_COLUMNS)) is not None:
        w.set_style(
            "grid_column_dsc_array",
            _grid_descriptor_array(f"{base_name}_column_dsc", columns),
        )
    if (align := layout.get(CONF_GRID_COLUMN_ALIGN)) is not None:
        w.set_style(CONF_GRID_COLUMN_ALIGN, literal(align))
    if (align := layout.get(CONF_GRID_ROW_ALIGN)) is not None:
        w.set_style(CONF_GRID_ROW_ALIGN, literal(align))
    if (flow := layout.get(CONF_FLEX_FLOW)) is not None:
        lv_obj.set_flex_flow(w.obj, literal(flow))
    if (main := layout.get(CONF_FLEX_ALIGN_MAIN)) is not None:
        w.set_style("flex_main_place", literal(main))
    if (cross := layout.get(CONF_FLEX_ALIGN_CROSS)) is not None:
        # Stretch is implemented at creation time by sizing the children; at
        # runtime we can only fall back to centering.
        if cross == "LV_FLEX_ALIGN_STRETCH":
            LOGGER.warning(
                "Flex cross alignment 'stretch' is not supported at runtime; using 'center' instead"
            )
            cross = "LV_FLEX_ALIGN_CENTER"
        w.set_style("flex_cross_place", literal(cross))
    if (track := layout.get(CONF_FLEX_ALIGN_TRACK)) is not None:
        w.set_style("flex_track_place", literal(track))


async def set_obj_properties(w: Widget, config):
    """Generate a list of C++ statements to apply properties to an lv_obj_t"""

    from ..schemas import ALL_STYLES, OBJ_PROPERTIES, remap_property

    if layout := config.get(CONF_LAYOUT):
        # The layout `type` (and the grid row/column structure) is only present
        # when a widget is created; update actions only change the layout style
        # options, leaving the type and grid structure unchanged.
        layout_type = layout.get(CONF_TYPE)
        if layout_type is not None:
            add_lv_use(layout_type)
            lv_obj.set_layout(w.obj, literal(f"LV_LAYOUT_{layout_type.upper()}"))
            # The widget's own id gives the grid descriptor arrays stable names.
            base_name = str(config[CONF_ID])
        else:
            base_name = None
        _set_layout_options(w, layout, base_name)
    parts = collect_parts(config)
    for part, states in parts.items():
        part = "LV_PART_" + part.upper()
        for state, props in states.items():
            state = "LV_STATE_" + state.upper()
            if state == "LV_STATE_DEFAULT":
                lv_state = literal(part)
            elif part == "LV_PART_MAIN":
                lv_state = literal(state)
            else:
                lv_state = join_enums((state, part))
            for style_id in props.get(CONF_STYLES, ()):
                w.add_style(style_id, lv_state)
            for prop, value in {
                k: v for k, v in props.items() if k in ALL_STYLES
            }.items():
                if isinstance(ALL_STYLES[prop], LValidator):
                    value = await ALL_STYLES[prop].process(value)
                prop_r = remap_property(prop)
                w.set_style(prop_r, value, lv_state)
    if group := config.get(CONF_GROUP):
        group = await cg.get_variable(group)
        lv.group_add_obj(group, w.obj)
    props = parts[CONF_MAIN][CONF_DEFAULT]
    lambs = {}
    flag_set = set()
    flag_clr = set()
    for prop, value in {k: v for k, v in props.items() if k in OBJ_FLAGS}.items():
        if isinstance(value, cv.Lambda):
            lambs[prop] = value
        elif value:
            flag_set.add(prop)
        else:
            flag_clr.add(prop)
    if flag_set:
        adds = join_enums(flag_set, "LV_OBJ_FLAG_")
        w.add_flag(adds)
    if flag_clr:
        clrs = join_enums(flag_clr, "LV_OBJ_FLAG_")
        w.clear_flag(clrs)
    for key, value in lambs.items():
        lamb = await cg.process_lambda(value, [], capture="=", return_type=cg.bool_)
        flag = f"LV_OBJ_FLAG_{key.upper()}"
        with LvConditional(call_lambda(lamb)) as cond:
            w.add_flag(flag)
            cond.else_()
            w.clear_flag(flag)

    for key, value in config.get(CONF_STATE, {}).items():
        if isinstance(value, cv.Lambda):
            value = call_lambda(
                await cg.process_lambda(value, [], capture="=", return_type=cg.bool_)
            )
        state = getattr(LV_STATE, key.upper())
        w.set_state(state, value)

    for property in OBJ_PROPERTIES:
        await w.set_property(property, config, lv_name="obj")


async def add_widgets(parent: Widget, config: dict):
    """
    Add all widgets to an object
    :param parent: The enclosing obj
    :param config: The configuration
    :return:
    """
    for w in config.get(CONF_WIDGETS, ()):
        w_type, w_cnfig = next(iter(w.items()))
        await widget_to_code(w_cnfig, w_type, parent.obj)


async def widget_to_code(w_cnfig, w_type: WidgetType | str, parent) -> Widget:
    """
    Converts a Widget definition to C code.
    :param w_cnfig: The widget configuration
    :param w_type:  The Widget type
    :param parent: The parent to which the widget should be added
    :return:
    """

    from ..schemas import WIDGET_TYPES

    spec: WidgetType = (
        w_type if isinstance(w_type, WidgetType) else WIDGET_TYPES[w_type]
    )
    return await spec.create_to_code(w_cnfig, parent)


class NumberType(WidgetType):
    async def get_max(self, config: dict):
        return await lv_int.process(config.get(CONF_MAX_VALUE, 100))

    async def get_min(self, config: dict):
        return await lv_int.process(config.get(CONF_MIN_VALUE, 0))



from esphome import automation
import esphome.codegen as cg
from esphome.components import sensor, uart
import esphome.config_validation as cv
from esphome.const import (
    CONF_BASELINE,
    CONF_CO2,
    CONF_ID,
    CONF_WARMUP_TIME,
    DEVICE_CLASS_CARBON_DIOXIDE,
    ICON_MOLECULE_CO2,
    STATE_CLASS_MEASUREMENT,
    UNIT_PARTS_PER_MILLION,
)

DEPENDENCIES = ["uart"]

hc8_ns = cg.esphome_ns.namespace("hc8")
HC8Component = hc8_ns.class_("HC8Component", cg.PollingComponent, uart.UARTDevice)
HC8CalibrateAction = hc8_ns.class_("HC8CalibrateAction", automation.Action)

CONFIG_SCHEMA = (
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(HC8Component),
            cv.Optional(CONF_CO2): sensor.sensor_schema(
                unit_of_measurement=UNIT_PARTS_PER_MILLION,
                icon=ICON_MOLECULE_CO2,
                accuracy_decimals=0,
                device_class=DEVICE_CLASS_CARBON_DIOXIDE,
                state_class=STATE_CLASS_MEASUREMENT,
            ),
            cv.Optional(
                CONF_WARMUP_TIME, default="75s"
            ): cv.positive_time_period_seconds,
        }
    )
    .extend(cv.polling_component_schema("60s"))
    .extend(uart.UART_DEVICE_SCHEMA)
)

FINAL_VALIDATE_SCHEMA = uart.final_validate_device_schema(
    "hc8",
    baud_rate=9600,
    require_rx=True,
    require_tx=True,
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    await uart.register_uart_device(var, config)

    if co2 := config.get(CONF_CO2):
        sens = await sensor.new_sensor(co2)
        cg.add(var.set_co2_sensor(sens))

    cg.add(var.set_warmup_seconds(config[CONF_WARMUP_TIME]))


CALIBRATION_ACTION_SCHEMA = cv.Schema(
    {
        cv.Required(CONF_ID): cv.use_id(HC8Component),
        cv.Required(CONF_BASELINE): cv.templatable(cv.uint16_t),
    }
)


@automation.register_action(
    "hc8.calibrate",
    HC8CalibrateAction,
    CALIBRATION_ACTION_SCHEMA,
    synchronous=True,
)
async def hc8_calibration_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    template_ = await cg.templatable(config[CONF_BASELINE], args, cg.uint16)
    cg.add(var.set_baseline(template_))
    return var



import esphome.codegen as cg
from esphome.components import esp32_ble_tracker
import esphome.config_validation as cv
from esphome.const import CONF_ID

DEPENDENCIES = ["esp32_ble_tracker"]

xiaomi_ble_ns = cg.esphome_ns.namespace("xiaomi_ble")
XiaomiListener = xiaomi_ble_ns.class_(
    "XiaomiListener", esp32_ble_tracker.ESPBTDeviceListener
)

CONFIG_SCHEMA = cv.Schema(
    {
        cv.GenerateID(): cv.declare_id(XiaomiListener),
    }
).extend(esp32_ble_tracker.ESP_BLE_DEVICE_SCHEMA)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await esp32_ble_tracker.register_ble_device(var, config)



import esphome.codegen as cg
from esphome.components import binary_sensor, rdm6300
import esphome.config_validation as cv
from esphome.const import CONF_UID

from . import rdm6300_ns

DEPENDENCIES = ["rdm6300"]

CONF_RDM6300_ID = "rdm6300_id"
RDM6300BinarySensor = rdm6300_ns.class_(
    "RDM6300BinarySensor", binary_sensor.BinarySensor
)

CONFIG_SCHEMA = binary_sensor.binary_sensor_schema(RDM6300BinarySensor).extend(
    {
        cv.GenerateID(CONF_RDM6300_ID): cv.use_id(rdm6300.RDM6300Component),
        cv.Required(CONF_UID): cv.uint32_t,
    }
)


async def to_code(config):
    var = await binary_sensor.new_binary_sensor(config)

    hub = await cg.get_variable(config[CONF_RDM6300_ID])
    cg.add(hub.register_card(var))
    cg.add(var.set_id(config[CONF_UID]))



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
import logging
from pathlib import Path

from PIL import Image, UnidentifiedImageError

import esphome.codegen as cg
from esphome.components.const import CONF_BYTE_ORDER, KEY_METADATA
import esphome.config_validation as cv
from esphome.const import CONF_DEFAULTS, CONF_FILE, CONF_ID, CONF_PLATFORM, CONF_TYPE
from esphome.core import CORE
from esphome.types import ConfigType

_LOGGER = logging.getLogger(__name__)

DOMAIN = "image"
DEPENDENCIES = ["display"]
IS_PLATFORM_COMPONENT = True

# Name of the built-in static-image platform (local file / web / MDI sources).
PLATFORM_FILE = "file"

image_ns = cg.esphome_ns.namespace("image")

ImageType = image_ns.enum("ImageType")


@dataclass(frozen=True)
class ImageMetaData:
    width: int
    height: int
    image_type: str
    transparency: str


CONF_OPAQUE = "opaque"
CONF_CHROMA_KEY = "chroma_key"
CONF_ALPHA_CHANNEL = "alpha_channel"
CONF_INVERT_ALPHA = "invert_alpha"
CONF_IMAGES = "images"

TRANSPARENCY_TYPES = (
    CONF_OPAQUE,
    CONF_CHROMA_KEY,
    CONF_ALPHA_CHANNEL,
)


def get_image_type_enum(type):
    return getattr(ImageType, f"IMAGE_TYPE_{type.upper()}")


def get_transparency_enum(transparency):
    return getattr(TransparencyType, f"TRANSPARENCY_{transparency.upper()}")


class ImageEncoder:
    """
    Superclass of image type encoders
    """

    # Control which transparency options are available for a given type
    allow_config = {CONF_ALPHA_CHANNEL, CONF_CHROMA_KEY, CONF_OPAQUE}

    # All imageencoder types are valid
    @staticmethod
    def validate(value):
        return value

    def __init__(self, width, height, transparency, dither, invert_alpha):
        """
        :param width:  The image width in pixels (or bytes)
        :param height:  The image height in pixels
        :param transparency: Transparency type
        :param dither: Dither method
        :param invert_alpha: True if the alpha channel should be inverted; for monochrome formats inverts the colours.
        """
        self.transparency = transparency
        self.width = width
        self.height = height
        self.data = [0] * width * height
        self.dither = dither
        self.index = 0
        self.invert_alpha = invert_alpha
        self.path = ""
        self.big_endian = False

    def convert(self, image, path):
        """
        Convert the image format
        :param image:  Input image
        :param path:  Path to the image file
        :return: converted image
        """
        return image

    def encode(self, pixel):
        """
        Encode a single pixel
        """

    def end_row(self):
        """
        Marks the end of a pixel row
        :return:
        """

    def end_image(self):
        """
        Called at the end of the image.
        :return:
        """

    def set_big_endian(self, big_endian: bool) -> None:
        self.big_endian = big_endian

    @classmethod
    def is_endian(cls) -> bool:
        """
        Check if the image encoder supports endianness configuration
        """
        return False


def is_alpha_only(image: Image):
    """
    Check if an image (assumed to be RGBA) is only alpha
    """
    # Any alpha data?
    if image.split()[-1].getextrema()[0] == 0xFF:
        return False
    return all(b.getextrema()[1] == 0 for b in image.split()[:-1])


class ImageBinary(ImageEncoder):
    allow_config = {CONF_OPAQUE, CONF_INVERT_ALPHA, CONF_CHROMA_KEY}

    def __init__(self, width, height, transparency, dither, invert_alpha):
        self.width8 = (width + 7) // 8
        super().__init__(self.width8, height, transparency, dither, invert_alpha)
        self.bitno = 0

    def convert(self, image, path):
        if is_alpha_only(image):
            image = image.split()[-1]
        return image.convert("1", dither=self.dither)

    def encode(self, pixel):
        if self.invert_alpha:
            pixel = not pixel
        if pixel:
            self.data[self.index] |= 0x80 >> (self.bitno % 8)
        self.bitno += 1
        if self.bitno == 8:
            self.bitno = 0
            self.index += 1

    def end_row(self):
        """
        Pad rows to a byte boundary
        """
        if self.bitno != 0:
            self.bitno = 0
            self.index += 1


class ImageGrayscale(ImageEncoder):
    allow_config = {CONF_ALPHA_CHANNEL, CONF_CHROMA_KEY, CONF_INVERT_ALPHA, CONF_OPAQUE}

    def convert(self, image, path):
        if is_alpha_only(image):
            if self.transparency != CONF_ALPHA_CHANNEL:
                _LOGGER.warning(
                    "Grayscale image %s is alpha only, but transparency is set to %s",
                    path,
                    self.transparency,
                )
                self.transparency = CONF_ALPHA_CHANNEL
            image = image.split()[-1]
        return image.convert("LA")

    def encode(self, pixel):
        b, a = pixel
        if self.transparency == CONF_CHROMA_KEY:
            if b == 1:
                b = 0
            if a != 0xFF:
                b = 1
        if self.invert_alpha:
            b ^= 0xFF
        if self.transparency == CONF_ALPHA_CHANNEL and a != 0xFF:
            b = a
        self.data[self.index] = b
        self.index += 1


class ImageRGB565(ImageEncoder):
    def __init__(self, width, height, transparency, dither, invert_alpha):
        super().__init__(
            width * 2,
            height,
            transparency,
            dither,
            invert_alpha,
        )
        self.alpha = [0] * width * height

    @classmethod
    def is_endian(cls) -> bool:
        """
        Check if the image encoder supports endianness configuration
        """
        return True

    def convert(self, image, path):
        return image.convert("RGBA")

    def encode(self, pixel):
        r, g, b, a = pixel
        r = r >> 3
        g = g >> 2
        b = b >> 3
        if self.invert_alpha:
            a ^= 0xFF
        self.alpha[self.index // 2] = a
        if self.transparency == CONF_CHROMA_KEY:
            if r == 0 and g == 1 and b == 0:
                g = 0
            elif a < 128:
                r = 0
                g = 1
                b = 0
        rgb = (r << 11) | (g << 5) | b
        if self.big_endian:
            self.data[self.index] = rgb >> 8
            self.index += 1
            self.data[self.index] = rgb & 0xFF
            self.index += 1
        else:
            self.data[self.index] = rgb & 0xFF
            self.index += 1
            self.data[self.index] = rgb >> 8
            self.index += 1

    def end_image(self):
        if self.transparency == CONF_ALPHA_CHANNEL:
            self.data.extend(self.alpha)


class ImageRGB(ImageEncoder):
    def __init__(self, width, height, transparency, dither, invert_alpha):
        stride = 4 if transparency == CONF_ALPHA_CHANNEL else 3
        super().__init__(
            width * stride,
            height,
            transparency,
            dither,
            invert_alpha,
        )

    def convert(self, image, path):
        return image.convert("RGBA")

    def encode(self, pixel):
        r, g, b, a = pixel
        if self.transparency == CONF_CHROMA_KEY:
            if r == 0 and g == 1 and b == 0:
                g = 0
            elif a < 128:
                r = 0
                g = 1
                b = 0
        self.data[self.index] = b
        self.index += 1
        self.data[self.index] = g
        self.index += 1
        self.data[self.index] = r
        self.index += 1
        if self.transparency == CONF_ALPHA_CHANNEL:
            if self.invert_alpha:
                a ^= 0xFF
            self.data[self.index] = a
            self.index += 1


class ReplaceWith:
    """
    Placeholder class to provide feedback on deprecated features
    """

    allow_config = {CONF_ALPHA_CHANNEL, CONF_CHROMA_KEY, CONF_OPAQUE}

    def __init__(self, replace_with):
        self.replace_with = replace_with

    def validate(self, value):
        raise cv.Invalid(
            f"Image type {value} is removed; replace with {self.replace_with}"
        )


IMAGE_TYPE = {
    "BINARY": ImageBinary,
    "GRAYSCALE": ImageGrayscale,
    "RGB565": ImageRGB565,
    "RGB": ImageRGB,
    "TRANSPARENT_BINARY": ReplaceWith("'type: BINARY' and 'transparency: chroma_key'"),
    "RGB24": ReplaceWith("'type: RGB'"),
    "RGBA": ReplaceWith("'type: RGB' and 'transparency: alpha_channel'"),
}

TransparencyType = image_ns.enum("TransparencyType")

CONF_TRANSPARENCY = "transparency"

Image_ = image_ns.class_("Image")

INSTANCE_TYPE = Image_


def is_svg_file(file):
    if not file:
        return False
    with Path(file).open("rb") as f:
        return "<svg" in str(f.read(1024))


def validate_transparency(choices=TRANSPARENCY_TYPES):
    def validate(value):
        if isinstance(value, bool):
            value = str(value)
        return cv.one_of(*choices, lower=True)(value)

    return validate


def validate_type(image_types):
    def validate(value):
        value = cv.one_of(*image_types, upper=True)(value)
        return IMAGE_TYPE[value].validate(value)

    return validate


def validate_settings(value, path=()):
    """
    Validate the settings for a single image configuration.
    """
    conf_type = value[CONF_TYPE]
    type_class = IMAGE_TYPE[conf_type]

    transparency = value.get(CONF_TRANSPARENCY, CONF_OPAQUE).lower()
    if transparency not in type_class.allow_config:
        raise cv.Invalid(
            f"Image format '{conf_type}' cannot have transparency: {transparency}"
        )
    invert_alpha = value.get(CONF_INVERT_ALPHA, False)
    if (
        invert_alpha
        and transparency != CONF_ALPHA_CHANNEL
        and CONF_INVERT_ALPHA not in type_class.allow_config
    ):
        raise cv.Invalid("No alpha channel to invert")
    if value.get(CONF_BYTE_ORDER) is not None and not type_class.is_endian():
        raise cv.Invalid(
            f"Image format '{conf_type}' does not support byte order configuration",
            path=path,
        )
    if file := value.get(CONF_FILE):
        file = Path(file)
        if not is_svg_file(file):
            try:
                Image.open(file)
            except UnidentifiedImageError as exc:
                raise cv.Invalid(
                    f"File can't be opened as image: {file.absolute()}", path=path
                ) from exc
    return value


def add_metadata(id: str, width: int, height: int, image_type: str, transparency):
    all_metadata = CORE.data.setdefault(DOMAIN, {}).setdefault(KEY_METADATA, {})
    all_metadata[str(id)] = ImageMetaData(
        width=width, height=height, image_type=image_type, transparency=transparency
    )


async def to_code(config: ConfigType) -> None:
    # Base platform-component codegen: each entry is generated by its platform's
    # own ``to_code``; here we only need the feature define to be present.
    cg.add_define("USE_IMAGE")


def get_all_image_metadata() -> dict[str, ImageMetaData]:
    """Get all image metadata."""
    return CORE.data.get(DOMAIN, {}).get(KEY_METADATA, {})


def get_image_metadata(image_id: str) -> ImageMetaData | None:
    """Get image metadata by ID for use by other components."""
    return get_all_image_metadata().get(image_id)


# ---------------------------------------------------------------------------
# Legacy top-level component -> `image:` platform deprecation helpers
# -- REMOVE after 2027.1.0 together with the `animation:`/`online_image:` shims.
#
# `animation:` and `online_image:` used to be standalone top-level components and
# are now platforms of `image:`. Their deprecated top-level shims use this helper
# to (1) record each raw entry as it is validated and (2) print a single,
# pasteable migrated `image:` block once every entry has been seen. The block is
# emitted from FINAL_VALIDATE_SCHEMA, which always runs after every per-entry
# CONFIG_SCHEMA step, so all entries are captured before it fires.
# ---------------------------------------------------------------------------


def legacy_platform_migration_warning(
    domain: str, platform: str, removal_version: str
) -> tuple[
    Callable[[ConfigType], ConfigType],
    Callable[[ConfigType], ConfigType],
]:
    """Build the per-entry capture and one-shot warning validators for a
    deprecated top-level component that is now an ``image:`` platform.

    Returns ``(capture, finalize)``:
    * ``capture`` is a ``CONFIG_SCHEMA`` validator placed *before* the real
      schema so it sees the raw user entry; it records a copy of each entry.
    * ``finalize`` is a ``FINAL_VALIDATE_SCHEMA`` validator that warns exactly
      once with the migrated, pasteable ``image:`` block.
    """
    entries_key = "legacy_entries"
    shown_key = "legacy_warning_shown"

    def capture(config: ConfigType) -> ConfigType:
        data = CORE.data.setdefault(domain, {})
        data.setdefault(entries_key, []).append(dict(config))
        return config

    def finalize(config: ConfigType) -> ConfigType:
        data = CORE.data.setdefault(domain, {})
        if not data.get(shown_key):
            data[shown_key] = True

            from esphome import yaml_util

            migrated = [
                {CONF_PLATFORM: platform, **entry}
                for entry in data.get(entries_key, [])
            ]
            _LOGGER.warning(
                "The top-level '%s:' configuration is deprecated and will be "
                "removed in ESPHome %s. '%s' is now a platform of the 'image' "
                "component. Replace your '%s:' block with:\n\n%s",
                domain,
                removal_version,
                domain,
                domain,
                yaml_util.dump({DOMAIN: migrated}),
            )
        return config

    return capture, finalize


# ---------------------------------------------------------------------------
# Legacy `image:` config migration -- REMOVE after 2027.1.0
#
# Before `image` became a platform component, its top-level config was either a
# bare list of image dicts, a single image dict, or a dict with `defaults:`,
# `images:` and per-type group keys. This block transparently rewrites those
# forms into the new ``platform: file`` list and prints the migrated YAML.
# It is intentionally self-contained so it can be deleted in one piece together
# with the ``LEGACY_CONFIG_MIGRATE`` assignment below.
# ---------------------------------------------------------------------------

LEGACY_REMOVAL_VERSION = "2027.1.0"


def _is_new_image_format(config: object) -> bool:
    """True when the config is already the new ``platform:``-tagged list."""
    return isinstance(config, list) and all(
        isinstance(entry, dict) and CONF_PLATFORM in entry for entry in config
    )


def _is_legacy_image_format(config: object) -> bool:
    """True when ``config`` matches a shape the pre-platform schema accepted.

    Only these shapes are migrated. Anything else -- a list containing a
    non-dict (or already platform-tagged) entry, or a dict with no recognised
    image keys -- is left untouched so the platform validation surfaces a
    proper error instead of the migration silently dropping the input.
    """
    if isinstance(config, list):
        # A bare list of (not-yet-platform-tagged) image dicts.
        return bool(config) and all(
            isinstance(entry, dict) and CONF_PLATFORM not in entry for entry in config
        )
    if not isinstance(config, dict):
        return False
    # A single image dict, or the grouped `defaults:`/`images:`/type-key form.
    return (
        CONF_ID in config
        or CONF_FILE in config
        or any(
            key in (CONF_DEFAULTS, CONF_IMAGES) or key.upper() in IMAGE_TYPE
            for key in config
        )
    )


def _flatten_legacy_image_config(config: object) -> list[dict]:
    """Structurally flatten a legacy ``image:`` config into image dicts.

    No validation or file IO is performed -- the ``file`` platform schema
    validates the resulting entries. Unrecognised shapes yield no entries so the
    normal platform validation surfaces the error.
    """
    if isinstance(config, list):
        return [dict(entry) for entry in config if isinstance(entry, dict)]
    if not isinstance(config, dict):
        return []
    if CONF_ID in config or CONF_FILE in config:
        return [dict(config)]

    defaults = config.get(CONF_DEFAULTS) or {}
    result: list[dict] = []

    def _add(entry: dict, extra: dict) -> None:
        merged = {**defaults, **extra, **entry}
        # The legacy `defaults:`/type-grouped forms only applied `byte_order` to
        # types that support it. Replicate that so an endian default merged into
        # e.g. a binary image stays valid.
        type_class = IMAGE_TYPE.get(str(merged.get(CONF_TYPE, "")).upper())
        if (
            CONF_BYTE_ORDER in merged
            and isinstance(type_class, type)
            and issubclass(type_class, ImageEncoder)
            and not type_class.is_endian()
        ):
            del merged[CONF_BYTE_ORDER]
        result.append(merged)

    def _add_entries(entries: object, extra: dict) -> None:
        # `entries` may be a single image dict or a list of them; non-dict
        # members are silently skipped, mirroring the old `ensure_list` leniency.
        for entry in [entries] if isinstance(entries, dict) else entries:
            if isinstance(entry, dict):
                _add(entry, extra)

    _add_entries(config.get(CONF_IMAGES, []), {})

    for key, value in config.items():
        if key in (CONF_DEFAULTS, CONF_IMAGES) or key.upper() not in IMAGE_TYPE:
            continue
        type_extra = {CONF_TYPE: key}
        if isinstance(value, dict) and (
            transparency_keys := [k for k in value if k in TRANSPARENCY_TYPES]
        ):
            for trans in transparency_keys:
                _add_entries(value[trans], {**type_extra, CONF_TRANSPARENCY: trans})
        elif isinstance(value, (list, dict)):
            _add_entries(value, type_extra)
    return result


def _migrate_legacy_image_config(config: object) -> list[dict] | None:
    """Rewrite a legacy ``image:`` config into the ``platform: file`` list.

    Returns None for the already-migrated platform form and for any shape the
    pre-platform schema never accepted, so normal platform validation can
    surface a proper error instead of the migration silently discarding input.
    """
    if _is_new_image_format(config) or not _is_legacy_image_format(config):
        return None
    migrated = [
        {CONF_PLATFORM: PLATFORM_FILE, **entry}
        for entry in _flatten_legacy_image_config(config)
    ]

    from esphome import yaml_util

    _LOGGER.warning(
        "The 'image:' configuration format is deprecated and will be removed in "
        "ESPHome %s. Images are now platforms of the 'image' component. Replace "
        "your 'image:' block with:\n\n%s",
        LEGACY_REMOVAL_VERSION,
        yaml_util.dump({DOMAIN: migrated}),
    )
    return migrated


LEGACY_CONFIG_MIGRATE = _migrate_legacy_image_config

# --------------------------- end legacy migration --------------------------



from esphome import automation
import esphome.codegen as cg
from esphome.components import display
import esphome.config_validation as cv
from esphome.const import (
    CONF_CALIBRATION,
    CONF_DISPLAY,
    CONF_MIRROR_X,
    CONF_MIRROR_Y,
    CONF_ON_RELEASE,
    CONF_ON_TOUCH,
    CONF_ON_UPDATE,
    CONF_SWAP_XY,
    CONF_TRANSFORM,
)
from esphome.core import CoroPriority, coroutine_with_priority

CODEOWNERS = ["@jesserockz", "@nielsnl68"]
DEPENDENCIES = ["display"]

IS_PLATFORM_COMPONENT = True

touchscreen_ns = cg.esphome_ns.namespace("touchscreen")

Touchscreen = touchscreen_ns.class_("Touchscreen", cg.PollingComponent)
TouchRotation = touchscreen_ns.enum("TouchRotation")
TouchPoint = touchscreen_ns.struct("TouchPoint")
TouchPoints_t = cg.std_vector.template(TouchPoint)
TouchPoints_t_const_ref = TouchPoints_t.operator("ref").operator("const")
TouchListener = touchscreen_ns.class_("TouchListener")

CONF_TOUCHSCREEN_ID = "touchscreen_id"
CONF_REPORT_INTERVAL = "report_interval"  # not used yet:
CONF_TOUCH_TIMEOUT = "touch_timeout"


CONF_X_MIN = "x_min"
CONF_X_MAX = "x_max"
CONF_Y_MIN = "y_min"
CONF_Y_MAX = "y_max"


def validate_calibration(calibration_config):
    x_min = calibration_config[CONF_X_MIN]
    x_max = calibration_config[CONF_X_MAX]
    y_min = calibration_config[CONF_Y_MIN]
    y_max = calibration_config[CONF_Y_MAX]
    if x_max < x_min:
        raise cv.Invalid(
            "x_min must be smaller than x_max. To mirror the direction use the 'transform' options"
        )
    if y_max < y_min:
        raise cv.Invalid(
            "y_min must be smaller than y_max. To mirror the direction use the 'transform' options"
        )
    x_delta = x_max - x_min
    y_delta = y_max - y_min
    if x_delta < 10 or y_delta < 10:
        raise cv.Invalid("Calibration value range must be greater than 10")
    return calibration_config


def option_with_default(option: str, defaults: dict, required: bool = False):
    if option in defaults or not required:
        return cv.Optional(option, default=defaults.get(option, cv.UNDEFINED))
    return cv.Required(option)


_CALIBRATION_KEYS = {CONF_X_MIN, CONF_X_MAX, CONF_Y_MIN, CONF_Y_MAX}
_TRANSFORM_KEYS = {CONF_SWAP_XY, CONF_MIRROR_X, CONF_MIRROR_Y}


def _calibration_schema(defaults: dict, required: bool) -> dict:
    """
    Generate Calibration schema. If defaults are provided for all suboptions,
    the entire calibration config is optional with a populated default value.
    Otherwise, it's optional or required as specified.
    """
    if _CALIBRATION_KEYS.issubset(defaults):
        key = cv.Optional(
            CONF_CALIBRATION,
            default={k: v for k, v in defaults.items() if k in _CALIBRATION_KEYS},
        )
    elif required:
        key = cv.Required(CONF_CALIBRATION)
    else:
        key = cv.Optional(CONF_CALIBRATION)
    return {
        key: cv.All(
            cv.Schema(
                {
                    option_with_default(x, defaults, True): cv.int_range(
                        min=0, max=4095
                    )
                    for x in _CALIBRATION_KEYS
                }
            ),
            validate_calibration,
        )
    }


def _transform_schema(defaults: dict) -> dict:
    if _TRANSFORM_KEYS.issubset(defaults):
        key = cv.Optional(
            CONF_TRANSFORM,
            default={k: v for k, v in defaults.items() if k in _TRANSFORM_KEYS},
        )
    else:
        key = cv.Optional(CONF_TRANSFORM)
    return {
        key: cv.Schema(
            {
                cv.Optional(x, default=defaults.get(x, False)): cv.boolean
                for x in _TRANSFORM_KEYS
            }
        )
    }


def touchscreen_schema(
    default_touch_timeout=cv.UNDEFINED,
    calibration_required=False,
    defaults: dict = None,
) -> cv.Schema:
    defaults = defaults or {}
    return cv.Schema(
        {
            cv.GenerateID(CONF_DISPLAY): cv.use_id(display.Display),
            cv.Optional(CONF_TOUCH_TIMEOUT, default=default_touch_timeout): cv.All(
                cv.positive_time_period_milliseconds,
                cv.Range(max=cv.TimePeriod(milliseconds=65535)),
            ),
            **_transform_schema(defaults),
            **_calibration_schema(defaults, calibration_required),
            cv.Optional(CONF_ON_TOUCH): automation.validate_automation(single=True),
            cv.Optional(CONF_ON_UPDATE): automation.validate_automation(single=True),
            cv.Optional(CONF_ON_RELEASE): automation.validate_automation(single=True),
        }
    ).extend(cv.polling_component_schema("50ms"))


TOUCHSCREEN_SCHEMA = touchscreen_schema(cv.UNDEFINED)


async def register_touchscreen(var, config):
    await cg.register_component(var, config)

    disp = await cg.get_variable(config[CONF_DISPLAY])
    cg.add(var.set_display(disp))

    if CONF_TOUCH_TIMEOUT in config:
        cg.add(var.set_touch_timeout(config[CONF_TOUCH_TIMEOUT]))

    if CONF_TRANSFORM in config:
        transform = config[CONF_TRANSFORM]
        cg.add(var.set_swap_xy(transform[CONF_SWAP_XY]))
        cg.add(var.set_mirror_x(transform[CONF_MIRROR_X]))
        cg.add(var.set_mirror_y(transform[CONF_MIRROR_Y]))

    if CONF_CALIBRATION in config:
        calibration_config = config[CONF_CALIBRATION]
        cg.add(
            var.set_calibration(
                calibration_config[CONF_X_MIN],
                calibration_config[CONF_X_MAX],
                calibration_config[CONF_Y_MIN],
                calibration_config[CONF_Y_MAX],
            )
        )

    if CONF_ON_TOUCH in config:
        await automation.build_automation(
            var.get_touch_trigger(),
            [(TouchPoint, "touch"), (TouchPoints_t_const_ref, "touches")],
            config[CONF_ON_TOUCH],
        )

    if CONF_ON_UPDATE in config:
        await automation.build_automation(
            var.get_update_trigger(),
            [(TouchPoints_t_const_ref, "touches")],
            config[CONF_ON_UPDATE],
        )

    if CONF_ON_RELEASE in config:
        await automation.build_automation(
            var.get_release_trigger(),
            [],
            config[CONF_ON_RELEASE],
        )


@coroutine_with_priority(CoroPriority.CORE)
async def to_code(config):
    cg.add_global(touchscreen_ns.using)
    cg.add_define("USE_TOUCHSCREEN")



# Async TCP client support for all platforms
import esphome.codegen as cg
import esphome.config_validation as cv
from esphome.core import CORE, CoroPriority, coroutine_with_priority

CODEOWNERS = ["@esphome/core"]
DEPENDENCIES = ["network"]


def AUTO_LOAD() -> list[str]:
    # Socket component needed for platforms using socket-based implementation
    # ESP32, ESP8266, RP2040, and LibreTiny use AsyncTCP libraries, others use sockets
    if (
        not CORE.is_esp32
        and not CORE.is_esp8266
        and not CORE.is_rp2
        and not CORE.is_libretiny
    ):
        return ["socket"]
    return []


# Support all platforms - Arduino/ESP-IDF get libraries, other platforms use socket implementation
CONFIG_SCHEMA = cv.Schema({})


@coroutine_with_priority(CoroPriority.NETWORK_TRANSPORT)
async def to_code(config):
    if CORE.is_esp32:
        # https://github.com/ESP32Async/AsyncTCP
        from esphome.components.esp32 import add_idf_component

        add_idf_component(name="esp32async/asynctcp", ref="3.4.91")
    elif CORE.is_libretiny:
        # https://github.com/ESP32Async/AsyncTCP
        cg.add_library("ESP32Async/AsyncTCP", "3.4.5")
    elif CORE.is_esp8266:
        # https://github.com/ESP32Async/ESPAsyncTCP
        cg.add_library("ESP32Async/ESPAsyncTCP", "2.0.0")
    elif CORE.is_rp2:
        # https://github.com/ayushsharma82/RPAsyncTCP
        # RPAsyncTCP is a drop-in replacement for AsyncTCP_RP2040W with better
        # ESPAsyncWebServer compatibility
        cg.add_library("ayushsharma82/RPAsyncTCP", "1.3.2")
    # Other platforms (host, etc) use socket-based implementation


def FILTER_SOURCE_FILES() -> list[str]:
    # Exclude socket implementation for platforms that use AsyncTCP libraries
    if CORE.is_esp32 or CORE.is_esp8266 or CORE.is_rp2 or CORE.is_libretiny:
        return ["async_tcp_socket.cpp"]
    return []



import esphome.codegen as cg
from esphome.components import sensor
import esphome.config_validation as cv
from esphome.const import (
    DEVICE_CLASS_PRESSURE,
    DEVICE_CLASS_TEMPERATURE,
    STATE_CLASS_MEASUREMENT,
    UNIT_CELSIUS,
    UNIT_REVOLUTIONS_PER_MINUTE,
)

from .. import (
    CONF_MICRONOVA_ID,
    MICRONOVA_ADDRESS_SCHEMA,
    MicroNova,
    MicroNovaListener,
    micronova_ns,
    to_code_micronova_listener,
)

UNIT_BAR = "bar"

MicroNovaSensor = micronova_ns.class_(
    "MicroNovaSensor", sensor.Sensor, MicroNovaListener
)

CONF_ROOM_TEMPERATURE = "room_temperature"
CONF_FUMES_TEMPERATURE = "fumes_temperature"
CONF_STOVE_POWER = "stove_power"
CONF_FAN_SPEED = "fan_speed"
CONF_WATER_TEMPERATURE = "water_temperature"
CONF_WATER_PRESSURE = "water_pressure"
CONF_MEMORY_ADDRESS_SENSOR = "memory_address_sensor"
CONF_FAN_RPM_OFFSET = "fan_rpm_offset"

CONFIG_SCHEMA = cv.Schema(
    {
        cv.GenerateID(CONF_MICRONOVA_ID): cv.use_id(MicroNova),
        cv.Optional(CONF_ROOM_TEMPERATURE): sensor.sensor_schema(
            MicroNovaSensor,
            unit_of_measurement=UNIT_CELSIUS,
            device_class=DEVICE_CLASS_TEMPERATURE,
            state_class=STATE_CLASS_MEASUREMENT,
            accuracy_decimals=1,
        ).extend(
            MICRONOVA_ADDRESS_SCHEMA(
                default_memory_location=0x00,
                default_memory_address=0x01,
                is_polling_component=True,
            )
        ),
        cv.Optional(CONF_FUMES_TEMPERATURE): sensor.sensor_schema(
            MicroNovaSensor,
            unit_of_measurement=UNIT_CELSIUS,
            device_class=DEVICE_CLASS_TEMPERATURE,
            state_class=STATE_CLASS_MEASUREMENT,
            accuracy_decimals=1,
        ).extend(
            MICRONOVA_ADDRESS_SCHEMA(
                default_memory_location=0x00,
                default_memory_address=0x5A,
                is_polling_component=True,
            )
        ),
        cv.Optional(CONF_STOVE_POWER): sensor.sensor_schema(
            MicroNovaSensor,
            state_class=STATE_CLASS_MEASUREMENT,
            accuracy_decimals=0,
        ).extend(
            MICRONOVA_ADDRESS_SCHEMA(
                default_memory_location=0x00,
                default_memory_address=0x34,
                is_polling_component=True,
            )
        ),
        cv.Optional(CONF_FAN_SPEED): sensor.sensor_schema(
            MicroNovaSensor,
            state_class=STATE_CLASS_MEASUREMENT,
            unit_of_measurement=UNIT_REVOLUTIONS_PER_MINUTE,
        )
        .extend(
            MICRONOVA_ADDRESS_SCHEMA(
                default_memory_location=0x00,
                default_memory_address=0x37,
                is_polling_component=True,
            )
        )
        .extend(
            {cv.Optional(CONF_FAN_RPM_OFFSET, default=0): cv.int_range(min=0, max=255)}
        ),
        cv.Optional(CONF_WATER_TEMPERATURE): sensor.sensor_schema(
            MicroNovaSensor,
            unit_of_measurement=UNIT_CELSIUS,
            device_class=DEVICE_CLASS_TEMPERATURE,
            state_class=STATE_CLASS_MEASUREMENT,
            accuracy_decimals=1,
        ).extend(
            MICRONOVA_ADDRESS_SCHEMA(
                default_memory_location=0x00,
                default_memory_address=0x3B,
                is_polling_component=True,
            )
        ),
        cv.Optional(CONF_WATER_PRESSURE): sensor.sensor_schema(
            MicroNovaSensor,
            unit_of_measurement=UNIT_BAR,
            device_class=DEVICE_CLASS_PRESSURE,
            state_class=STATE_CLASS_MEASUREMENT,
            accuracy_decimals=1,
        ).extend(
            MICRONOVA_ADDRESS_SCHEMA(
                default_memory_location=0x00,
                default_memory_address=0x3C,
                is_polling_component=True,
            )
        ),
        cv.Optional(CONF_MEMORY_ADDRESS_SENSOR): sensor.sensor_schema(
            MicroNovaSensor,
        ).extend(
            MICRONOVA_ADDRESS_SCHEMA(
                is_polling_component=True,
            )
        ),
    }
)


async def to_code(config):
    mv = await cg.get_variable(config[CONF_MICRONOVA_ID])

    for key, divisor in {
        CONF_ROOM_TEMPERATURE: 2,
        CONF_FUMES_TEMPERATURE: None,
        CONF_STOVE_POWER: None,
        CONF_MEMORY_ADDRESS_SENSOR: None,
        CONF_WATER_TEMPERATURE: 2,
        CONF_WATER_PRESSURE: 10,
    }.items():
        if sensor_config := config.get(key):
            sens = await sensor.new_sensor(sensor_config, mv)
            await to_code_micronova_listener(mv, sens, sensor_config)
            if divisor:
                cg.add(sens.set_divisor(divisor))

    if fan_speed_config := config.get(CONF_FAN_SPEED):
        sens = await sensor.new_sensor(fan_speed_config, mv)
        await to_code_micronova_listener(mv, sens, fan_speed_config)
        cg.add(sens.set_fan_speed_offset(fan_speed_config[CONF_FAN_RPM_OFFSET]))



from esphome import automation
import esphome.codegen as cg
from esphome.components import i2c
from esphome.components.audio_dac import AudioDac
import esphome.config_validation as cv
from esphome.const import CONF_ID, CONF_MODE

CODEOWNERS = ["@kbx81"]
DEPENDENCIES = ["i2c"]

aic3204_ns = cg.esphome_ns.namespace("aic3204")
AIC3204 = aic3204_ns.class_("AIC3204", AudioDac, cg.Component, i2c.I2CDevice)

SetAutoMuteAction = aic3204_ns.class_("SetAutoMuteAction", automation.Action)

CONFIG_SCHEMA = (
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(AIC3204),
        }
    )
    .extend(cv.COMPONENT_SCHEMA)
    .extend(i2c.i2c_device_schema(0x18))
)


SET_AUTO_MUTE_ACTION_SCHEMA = cv.maybe_simple_value(
    {
        cv.GenerateID(): cv.use_id(AIC3204),
        cv.Required(CONF_MODE): cv.templatable(cv.int_range(max=7, min=0)),
    },
    key=CONF_MODE,
)


@automation.register_action(
    "aic3204.set_auto_mute_mode",
    SetAutoMuteAction,
    SET_AUTO_MUTE_ACTION_SCHEMA,
    synchronous=True,
)
async def aic3204_set_volume_to_code(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    var = cg.new_Pvariable(action_id, template_arg, paren)

    template_ = await cg.templatable(config.get(CONF_MODE), args, cg.uint8)
    cg.add(var.set_auto_mute_mode(template_))

    return var


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    await i2c.register_i2c_device(var, config)



import esphome.config_validation as cv
from esphome.const import CONF_BRIGHTNESS, CONF_COLOR, CONF_LED

from ..defines import CONF_MAIN
from ..lv_validation import lv_color, lv_percentage
from ..lvcode import lv
from ..types import LvType
from . import Widget, WidgetType

LED_SCHEMA = cv.Schema(
    {
        cv.Optional(CONF_COLOR): lv_color,
        cv.Optional(CONF_BRIGHTNESS): lv_percentage,
    }
)


class LedType(WidgetType):
    def __init__(self):
        super().__init__(CONF_LED, LvType("lv_led_t"), (CONF_MAIN,), LED_SCHEMA)

    async def to_code(self, w: Widget, config):
        if (color := config.get(CONF_COLOR)) is not None:
            lv.led_set_color(w.obj, await lv_color.process(color))
        if (brightness := config.get(CONF_BRIGHTNESS)) is not None:
            lv.led_set_brightness(w.obj, await lv_percentage.process(brightness))


led_spec = LedType()



import logging
from typing import Any

import esphome.config_validation as cv
from esphome.const import (
    CONF_DISABLED,
    CONF_INPUT,
    CONF_MODE,
    CONF_NUMBER,
    PLATFORM_ESP32,
)
from esphome.pins import PIN_SCHEMA_REGISTRY, check_strapping_pin
from esphome.types import ConfigType

_ESP32S3_SPI_PSRAM_PINS = {
    26: "SPICS1",
    27: "SPIHD",
    28: "SPIWP",
    29: "SPICS0",
    30: "SPICLK",
    31: "SPIQ",
    32: "SPID",
}

_ESP32S3R8_PSRAM_PINS = {
    33: "SPIIO4",
    34: "SPIIO5",
    35: "SPIIO6",
    36: "SPIIO7",
    37: "SPIDQS",
}

_ESP32S3_USB_JTAG_PINS = {19, 20}

_ESP32S3_STRAPPING_PINS = {0, 3, 45, 46}

_LOGGER = logging.getLogger(__name__)


def esp32_s3_validate_gpio_pin(value: int) -> int:
    if value < 0 or value > 48:
        raise cv.Invalid(f"Invalid pin number: {value} (must be 0-48)")

    if value in _ESP32S3_SPI_PSRAM_PINS:
        raise cv.Invalid(
            f"This pin cannot be used on ESP32-S3s and is already used by the SPI/PSRAM interface(function: {_ESP32S3_SPI_PSRAM_PINS[value]})"
        )
    # GPIO33-37 (_ESP32S3R8_PSRAM_PINS) are only taken by the PSRAM interface in
    # octal mode -- whether that applies isn't known here, so the warning is
    # deferred to final_validate_pins() in gpio.py once the PSRAM mode is resolved.

    if value in (22, 23, 24, 25):
        # These pins are not exposed in GPIO mux (reason unknown)
        # but they're missing from IO_MUX list in datasheet
        raise cv.Invalid(f"The pin GPIO{value} is not usable on ESP32-S3s.")
    if value in _ESP32S3_USB_JTAG_PINS:
        _LOGGER.warning(
            "GPIO%d is used by the USB-Serial-JTAG interface."
            " Using this pin as GPIO will conflict with USB-Serial-JTAG.",
            value,
        )

    return value


def esp32_s3_validate_supports(value: dict[str, Any]) -> dict[str, Any]:
    num = value[CONF_NUMBER]
    mode = value[CONF_MODE]
    is_input = mode[CONF_INPUT]

    if num < 0 or num > 48:
        raise cv.Invalid(f"Invalid pin number: {num} (must be 0-48)")
    if is_input:
        # All ESP32 pins support input mode
        pass

    check_strapping_pin(value, _ESP32S3_STRAPPING_PINS, _LOGGER)
    return value


def esp32_s3_final_validate_pins(full_config: ConfigType) -> None:
    """Warn about GPIO33-37 usage, but only when octal PSRAM (which uses them) is set.

    These pins are only taken by the PSRAM interface in octal mode (ESP32-S3R8 /
    S3R8V); on quad-PSRAM variants -- or when the psram block is disabled, so the
    octal interface is never configured -- they are free. The per-pin validator
    can't know the PSRAM mode, so the check is deferred here, where
    PIN_SCHEMA_REGISTRY.pins_used already lists every used pin.
    """
    # Imported locally to avoid circular import issues
    from esphome.components.psram import DOMAIN as PSRAM_DOMAIN, TYPE_OCTAL

    psram_config = full_config.get(PSRAM_DOMAIN, {})
    if psram_config.get(CONF_DISABLED) or psram_config.get(CONF_MODE) != TYPE_OCTAL:
        return
    for number in sorted(
        number
        for key, _client_id, number in PIN_SCHEMA_REGISTRY.pins_used
        if key == PLATFORM_ESP32 and number in _ESP32S3R8_PSRAM_PINS
    ):
        _LOGGER.warning(
            "GPIO%d is used by the PSRAM interface in octal mode and should be avoided",
            number,
        )



import esphome.codegen as cg
from esphome.components import output
import esphome.config_validation as cv
from esphome.const import CONF_CHANNEL, CONF_ID

from .. import TLC5947, tlc5947_ns

DEPENDENCIES = ["tlc5947"]

TLC5947Channel = tlc5947_ns.class_(
    "TLC5947Channel", output.FloatOutput, cg.Parented.template(TLC5947)
)

CONF_TLC5947_ID = "tlc5947_id"
CONFIG_SCHEMA = output.FLOAT_OUTPUT_SCHEMA.extend(
    {
        cv.GenerateID(CONF_TLC5947_ID): cv.use_id(TLC5947),
        cv.Required(CONF_ID): cv.declare_id(TLC5947Channel),
        cv.Required(CONF_CHANNEL): cv.uint16_t,
    }
).extend(cv.COMPONENT_SCHEMA)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await output.register_output(var, config)
    await cg.register_parented(var, config[CONF_TLC5947_ID])
    cg.add(var.set_channel(config[CONF_CHANNEL]))



from esphome import pins
import esphome.codegen as cg
from esphome.components import switch
import esphome.config_validation as cv
from esphome.const import CONF_OPTIMISTIC, CONF_PULSE_LENGTH, CONF_WAIT_TIME

from .. import hbridge_ns

HBridgeSwitch = hbridge_ns.class_("HBridgeSwitch", switch.Switch, cg.Component)

CODEOWNERS = ["@dwmw2"]

CONF_OFF_PIN = "off_pin"
CONF_ON_PIN = "on_pin"

CONFIG_SCHEMA = (
    switch.switch_schema(HBridgeSwitch)
    .extend(
        {
            cv.Required(CONF_ON_PIN): pins.gpio_output_pin_schema,
            cv.Required(CONF_OFF_PIN): pins.gpio_output_pin_schema,
            cv.Optional(
                CONF_PULSE_LENGTH, default="100ms"
            ): cv.positive_time_period_milliseconds,
            cv.Optional(CONF_WAIT_TIME): cv.positive_time_period_milliseconds,
            cv.Optional(CONF_OPTIMISTIC, default=False): cv.boolean,
        }
    )
    .extend(cv.COMPONENT_SCHEMA)
)


async def to_code(config):
    var = await switch.new_switch(config)
    await cg.register_component(var, config)

    on_pin = await cg.gpio_pin_expression(config[CONF_ON_PIN])
    cg.add(var.set_on_pin(on_pin))
    off_pin = await cg.gpio_pin_expression(config[CONF_OFF_PIN])
    cg.add(var.set_off_pin(off_pin))
    cg.add(var.set_pulse_length(config[CONF_PULSE_LENGTH]))
    cg.add(var.set_optimistic(config[CONF_OPTIMISTIC]))
    if wait_time := config.get(CONF_WAIT_TIME):
        cg.add(var.set_wait_time(wait_time))



from esphome import automation
import esphome.codegen as cg
from esphome.components import ble_client, esp32_ble_tracker, sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_CHARACTERISTIC_UUID,
    CONF_LAMBDA,
    CONF_NOTIFY,
    CONF_SERVICE_UUID,
    CONF_TRIGGER_ID,
    CONF_TYPE,
    DEVICE_CLASS_SIGNAL_STRENGTH,
    STATE_CLASS_MEASUREMENT,
    UNIT_DECIBEL_MILLIWATT,
)

from .. import ble_client_ns

DEPENDENCIES = ["ble_client"]

CONF_DESCRIPTOR_UUID = "descriptor_uuid"

CONF_ON_NOTIFY = "on_notify"
TYPE_CHARACTERISTIC = "characteristic"
TYPE_RSSI = "rssi"

adv_data_t = cg.std_vector.template(cg.uint8)
adv_data_t_const_ref = adv_data_t.operator("ref").operator("const")

BLESensor = ble_client_ns.class_(
    "BLESensor", sensor.Sensor, cg.PollingComponent, ble_client.BLEClientNode
)
BLESensorNotifyTrigger = ble_client_ns.class_(
    "BLESensorNotifyTrigger", automation.Trigger.template(cg.float_)
)

BLEClientRssiSensor = ble_client_ns.class_(
    "BLEClientRSSISensor", sensor.Sensor, cg.PollingComponent, ble_client.BLEClientNode
)


def checkType(value):
    if CONF_TYPE not in value and CONF_SERVICE_UUID in value:
        raise cv.Invalid(
            "Looks like you're trying to create a ble characteristic sensor. Please add `type: characteristic` to your sensor config."
        )
    return value


CONFIG_SCHEMA = cv.All(
    checkType,
    cv.typed_schema(
        {
            TYPE_CHARACTERISTIC: sensor.sensor_schema(
                BLESensor,
                accuracy_decimals=0,
            )
            .extend(cv.polling_component_schema("60s"))
            .extend(ble_client.BLE_CLIENT_SCHEMA)
            .extend(
                {
                    cv.Required(CONF_SERVICE_UUID): esp32_ble_tracker.bt_uuid,
                    cv.Required(CONF_CHARACTERISTIC_UUID): esp32_ble_tracker.bt_uuid,
                    cv.Optional(CONF_DESCRIPTOR_UUID): esp32_ble_tracker.bt_uuid,
                    cv.Optional(CONF_LAMBDA): cv.returning_lambda,
                    cv.Optional(CONF_NOTIFY, default=False): cv.boolean,
                    cv.Optional(CONF_ON_NOTIFY): automation.validate_automation(
                        {
                            cv.GenerateID(CONF_TRIGGER_ID): cv.declare_id(
                                BLESensorNotifyTrigger
                            ),
                        }
                    ),
                }
            ),
            TYPE_RSSI: sensor.sensor_schema(
                BLEClientRssiSensor,
                accuracy_decimals=0,
                unit_of_measurement=UNIT_DECIBEL_MILLIWATT,
                device_class=DEVICE_CLASS_SIGNAL_STRENGTH,
                state_class=STATE_CLASS_MEASUREMENT,
            )
            .extend(cv.polling_component_schema("60s"))
            .extend(ble_client.BLE_CLIENT_SCHEMA),
        },
        lower=True,
    ),
)


async def rssi_sensor_to_code(config):
    var = await sensor.new_sensor(config)
    await cg.register_component(var, config)
    await ble_client.register_ble_node(var, config)


async def characteristic_sensor_to_code(config):
    var = await sensor.new_sensor(config)
    if len(config[CONF_SERVICE_UUID]) == len(esp32_ble_tracker.bt_uuid16_format):
        cg.add(
            var.set_service_uuid16(esp32_ble_tracker.as_hex(config[CONF_SERVICE_UUID]))
        )
    elif len(config[CONF_SERVICE_UUID]) == len(esp32_ble_tracker.bt_uuid32_format):
        cg.add(
            var.set_service_uuid32(esp32_ble_tracker.as_hex(config[CONF_SERVICE_UUID]))
        )
    elif len(config[CONF_SERVICE_UUID]) == len(esp32_ble_tracker.bt_uuid128_format):
        uuid128 = esp32_ble_tracker.as_reversed_hex_array(config[CONF_SERVICE_UUID])
        cg.add(var.set_service_uuid128(uuid128))

    if len(config[CONF_CHARACTERISTIC_UUID]) == len(esp32_ble_tracker.bt_uuid16_format):
        cg.add(
            var.set_char_uuid16(
                esp32_ble_tracker.as_hex(config[CONF_CHARACTERISTIC_UUID])
            )
        )
    elif len(config[CONF_CHARACTERISTIC_UUID]) == len(
        esp32_ble_tracker.bt_uuid32_format
    ):
        cg.add(
            var.set_char_uuid32(
                esp32_ble_tracker.as_hex(config[CONF_CHARACTERISTIC_UUID])
            )
        )
    elif len(config[CONF_CHARACTERISTIC_UUID]) == len(
        esp32_ble_tracker.bt_uuid128_format
    ):
        uuid128 = esp32_ble_tracker.as_reversed_hex_array(
            config[CONF_CHARACTERISTIC_UUID]
        )
        cg.add(var.set_char_uuid128(uuid128))

    if descriptor_uuid := config.get(CONF_DESCRIPTOR_UUID):
        if len(descriptor_uuid) == len(esp32_ble_tracker.bt_uuid16_format):
            cg.add(var.set_descr_uuid16(esp32_ble_tracker.as_hex(descriptor_uuid)))
        elif len(descriptor_uuid) == len(esp32_ble_tracker.bt_uuid32_format):
            cg.add(var.set_descr_uuid32(esp32_ble_tracker.as_hex(descriptor_uuid)))
        elif len(descriptor_uuid) == len(esp32_ble_tracker.bt_uuid128_format):
            uuid128 = esp32_ble_tracker.as_reversed_hex_array(descriptor_uuid)
            cg.add(var.set_descr_uuid128(uuid128))

    if lambda_config := config.get(CONF_LAMBDA):
        lambda_ = await cg.process_lambda(
            lambda_config, [(adv_data_t_const_ref, "x")], return_type=cg.float_
        )
        cg.add(var.set_data_to_value(lambda_))

    await cg.register_component(var, config)
    await ble_client.register_ble_node(var, config)
    cg.add(var.set_enable_notify(config[CONF_NOTIFY]))
    for conf in config.get(CONF_ON_NOTIFY, []):
        trigger = cg.new_Pvariable(conf[CONF_TRIGGER_ID], var)
        await ble_client.register_ble_node(trigger, config)
        await automation.build_automation(trigger, [(float, "x")], conf)


async def to_code(config):
    if config[CONF_TYPE] == TYPE_RSSI:
        await rssi_sensor_to_code(config)
    elif config[CONF_TYPE] == TYPE_CHARACTERISTIC:
        await characteristic_sensor_to_code(config)



import esphome.codegen as cg
from esphome.components import select
import esphome.config_validation as cv
from esphome.const import (
    CONF_ENUM_DATAPOINT,
    CONF_INT_DATAPOINT,
    CONF_OPTIMISTIC,
    CONF_OPTIONS,
)

from .. import CONF_TUYA_ID, Tuya, tuya_ns

DEPENDENCIES = ["tuya"]
CODEOWNERS = ["@bearpawmaxim"]

TuyaSelect = tuya_ns.class_("TuyaSelect", select.Select, cg.Component)


def ensure_option_map(value):
    cv.check_not_templatable(value)
    option = cv.All(cv.int_range(0, 2**8 - 1))
    mapping = cv.All(cv.string_strict)
    options_map_schema = cv.Schema({option: mapping})
    value = options_map_schema(value)

    all_values = list(value.keys())
    unique_values = set(value.keys())
    if len(all_values) != len(unique_values):
        raise cv.Invalid("Mapping values must be unique.")

    return value


CONFIG_SCHEMA = cv.All(
    select.select_schema(TuyaSelect)
    .extend(
        {
            cv.GenerateID(CONF_TUYA_ID): cv.use_id(Tuya),
            cv.Optional(CONF_ENUM_DATAPOINT): cv.uint8_t,
            cv.Optional(CONF_INT_DATAPOINT): cv.uint8_t,
            cv.Required(CONF_OPTIONS): ensure_option_map,
            cv.Optional(CONF_OPTIMISTIC, default=False): cv.boolean,
        }
    )
    .extend(cv.COMPONENT_SCHEMA),
    cv.has_exactly_one_key(CONF_ENUM_DATAPOINT, CONF_INT_DATAPOINT),
)


async def to_code(config):
    options_map = config[CONF_OPTIONS]
    var = await select.new_select(config, options=list(options_map.values()))
    await cg.register_component(var, config)
    cg.add(var.set_select_mappings(list(options_map.keys())))
    parent = await cg.get_variable(config[CONF_TUYA_ID])
    cg.add(var.set_tuya_parent(parent))
    if (enum_datapoint := config.get(CONF_ENUM_DATAPOINT, None)) is not None:
        cg.add(var.set_select_id(enum_datapoint, False))
    if (int_datapoint := config.get(CONF_INT_DATAPOINT, None)) is not None:
        cg.add(var.set_select_id(int_datapoint, True))
    cg.add(var.set_optimistic(config[CONF_OPTIMISTIC]))



import functools
import importlib
from pathlib import Path
import pkgutil
import re

from esphome.automation import Trigger, build_automation, validate_automation
import esphome.codegen as cg
from esphome.components.const import (
    BYTE_ORDER_BIG,
    CONF_BYTE_ORDER,
    CONF_COLOR_DEPTH,
    CONF_DRAW_ROUNDING,
)
from esphome.components.display import Display, get_display_metadata, validate_rotation
from esphome.components.esp32 import (
    VARIANT_ESP32P4,
    add_idf_component,
    add_idf_sdkconfig_option,
    get_esp32_variant,
)
from esphome.components.image import (
    CONF_OPAQUE,
    IMAGE_TYPE,
    ImageBinary,
    ImageGrayscale,
    ImageRGB,
    ImageRGB565,
    get_image_metadata,
)
from esphome.components.psram import DOMAIN as PSRAM_DOMAIN
import esphome.config_validation as cv
from esphome.const import (
    CONF_BUFFER_SIZE,
    CONF_ESPHOME,
    CONF_GROUP,
    CONF_ID,
    CONF_LOG_LEVEL,
    CONF_ON_IDLE,
    CONF_PAGES,
    CONF_PLATFORMIO_OPTIONS,
    CONF_ROTATION,
    CONF_TIMEOUT,
    CONF_TRIGGER_ID,
)
from esphome.core import CORE, ID, Lambda
from esphome.cpp_generator import MockObj
from esphome.final_validate import full_config
from esphome.helpers import write_file_if_changed
from esphome.schema_extractors import SCHEMA_EXTRACT, schema_extractor
from esphome.writer import clean_build
from esphome.yaml_util import load_yaml

from . import defines as df, lv_validation as lvalid, widgets
from .animation import ANIMATION_SCHEMA, add_animation_triggers, animations_to_code
from .automation import layers_to_code, lvgl_update
from .defines import (
    CONF_ALIGN_TO_LAMBDA_ID,
    CONF_ANIMATIONS,
    LOGGER,
    add_lv_use,
    get_focused_widgets,
    get_lv_images_used,
    get_refreshed_widgets,
    set_widgets_completed,
)
from .encoders import (
    ENCODERS_CONFIG,
    encoders_to_code,
    get_default_group,
    initial_focus_to_code,
)
from .gradient import GRADIENT_SCHEMA, gradients_to_code
from .keypads import KEYPADS_CONFIG, keypads_to_code
from .lv_validation import lv_bool
from .lvcode import LvContext, LvglComponent, lv_event_t_ptr, lvgl_static
from .schemas import (
    BASE_PROPS,
    DISP_BG_SCHEMA,
    FULL_STYLE_SCHEMA,
    SET_STATE_SCHEMA,
    STATE_SCHEMA,
    STYLE_REMAP,
    STYLE_SCHEMA,
    WIDGET_TYPES,
    any_widget_schema,
    container_schema,
    container_schema_value,
    obj_dict,
)
from .styles import styles_to_code, theme_to_code
from .touchscreens import touchscreen_schema, touchscreens_to_code
from .trigger import generate_align_tos, generate_triggers
from .types import (
    IdleTrigger,
    PlainTrigger,
    RotationType,
    lv_font_t,
    lv_group_t,
    lv_lambda_t,
    lv_obj_t_ptr,
    lv_style_t,
    lvgl_ns,
)
from .widgets import (
    LvScrActType,
    Widget,
    add_widgets,
    get_screen_active,
    set_obj_properties,
)
from .widgets.img import CONF_IMAGE

# Import only what we actually use directly in this file
from .widgets.msgbox import MSGBOX_SCHEMA, msgboxes_to_code
from .widgets.obj import obj_spec  # Used in LVGL_SCHEMA
from .widgets.page import (  # page_spec used in LVGL_SCHEMA
    add_pages,
    generate_page_triggers,
    page_spec,
)

# These style schemas live in .schemas but are imported here so they land in
# this module's namespace, where script/build_language_schema.py registers them
# as *named* schemas and emits `extends` references — instead of inlining the
# ~80-property STYLE_SCHEMA at every widget x part x state, which bloated the
# dumped lvgl schema ~23x (17 MB vs ~750 KB). They are not otherwise used in
# this file; this tuple keeps the imports live (and self-documents why).
_SCHEMA_DUMPER_NAMED_SCHEMAS = (STYLE_SCHEMA, STATE_SCHEMA, SET_STATE_SCHEMA)

# Widget registration happens via WidgetType.__init__ in individual widget files
# The imports below trigger creation of the widget types
# Action registration (lvgl.{widget}.update) happens automatically
# in the WidgetType.__init__ method

for module_info in pkgutil.iter_modules(widgets.__path__):
    importlib.import_module(f".widgets.{module_info.name}", package=__package__)

DOMAIN = "lvgl"
DEPENDENCIES = ["display"]
AUTO_LOAD = ["key_provider"]
CODEOWNERS = ["@clydebarrow"]
HELLO_WORLD_FILE = "hello_world.yaml"


SIMPLE_TRIGGERS = (
    df.CONF_ON_PAUSE,
    df.CONF_ON_RESUME,
    df.CONF_ON_DRAW_START,
    df.CONF_ON_DRAW_END,
    df.CONF_ON_LANDSCAPE,
    df.CONF_ON_PORTRAIT,
)


def as_macro(macro, value):
    if value is None:
        return f"#define {macro}"
    return f"#define {macro} {value}"


LVGL_VERSION = "9.5.0"
LV_CONF_FILENAME = "lv_conf.h"
LV_CONF_H_FORMAT = """\
#pragma once
{}
"""


def generate_lv_conf_h():
    # Get all possible LV_ config defines based on the widgets used in the config, and the standard LVGL options
    all_defines = set(
        df.LV_DEFINES + tuple(f"LV_USE_{w.upper()}" for w in WIDGET_TYPES)
    )
    build_flags = (
        CORE.config[CONF_ESPHOME].get(CONF_PLATFORMIO_OPTIONS).get("build_flags", [])
    )
    if not isinstance(build_flags, list):
        build_flags = [build_flags]
    # Extract define names from build flags like '-DLV_USE_CHART=1', '-D LV_USE_CHART',
    # or multiple defines in one string.
    define_pattern = r'-D\s*([A-Z_][A-Z0-9_]*)(?:=[^\s\'"\]]*)?'
    defines_from_flags = {
        m.group(1) for flag in build_flags for m in re.finditer(define_pattern, flag)
    }

    # Get the defines that are actually used based on the config,
    lv_defines = df.get_defines()
    clashes = defines_from_flags & lv_defines.keys()
    if clashes:
        LOGGER.warning(
            "Some defines are set both by ESPHome build flags and by LVGL configuration which may lead to unexpected behavior: %s",
            sorted(clashes),
        )
    unused_defines = all_defines - lv_defines.keys() - defines_from_flags

    # Create the content of lv_conf.h with the used defines set to their value, and the unused defines disabled
    definitions = [as_macro(m, v) for m, v in lv_defines.items()] + [
        as_macro(m, "0") for m in unused_defines
    ]
    definitions.sort()
    return LV_CONF_H_FORMAT.format("\n".join(definitions))


def multi_conf_validate(configs: list[dict]):
    displays = [config[df.CONF_DISPLAYS] for config in configs]
    # flatten the display list
    display_list = [disp for disps in displays for disp in disps]
    if len(display_list) != len(set(display_list)):
        raise cv.Invalid("A display ID may be used in only one LVGL instance")
    for config in configs:
        for item in (df.CONF_ENCODERS, df.CONF_KEYPADS):
            for enc in config.get(item, ()):
                if CONF_GROUP not in enc:
                    raise cv.Invalid(
                        f"'{item}' must have an explicit group set when using multiple LVGL instances"
                    )
    base_config = configs[0]
    for config in configs[1:]:
        for item in (
            CONF_LOG_LEVEL,
            CONF_COLOR_DEPTH,
            CONF_BYTE_ORDER,
            df.CONF_TRANSPARENCY_KEY,
        ):
            if base_config[item] != config[item]:
                raise cv.Invalid(
                    f"Config item '{item}' must be the same for all LVGL instances"
                )


def final_validation(config_list):
    global_config = full_config.get()
    # Resolve byte_order from display metadata before multi-config validation
    for config in config_list:
        metas = [get_display_metadata(disp) for disp in config[df.CONF_DISPLAYS]]
        if any(m.has_writer for m in metas):
            raise cv.Invalid(
                "Using lambda:, pages:, auto_clear_enabled: true, or show_test_card: true in display config is not compatible with LVGL"
            )
        if any(m.rotation != 0 for m in metas):
            raise cv.Invalid(
                "use of 'rotation' in the display config is not compatible with LVGL, please set rotation in the LVGL config instead"
            )
        config[CONF_DRAW_ROUNDING] = max(
            [m.draw_rounding for m in metas] + [config[CONF_DRAW_ROUNDING]]
        )
        display_byte_orders = {
            m.byte_order for m in metas if m.byte_order is not cv.UNDEFINED
        }
        if len(display_byte_orders) > 1:
            raise cv.Invalid(
                "All displays configured for an LVGL instance must use the same byte_order"
            )
        if display_byte_orders:
            display_order = next(iter(display_byte_orders))
            if CONF_BYTE_ORDER in config:
                if config[CONF_BYTE_ORDER] != display_order:
                    raise cv.Invalid(
                        "LVGL byte order must match the display byte order",
                        [CONF_BYTE_ORDER],
                    )
            else:
                config[CONF_BYTE_ORDER] = display_order
        if CONF_BYTE_ORDER not in config:
            config[CONF_BYTE_ORDER] = BYTE_ORDER_BIG

        if (pages := config.get(CONF_PAGES)) and all(p[df.CONF_SKIP] for p in pages):
            raise cv.Invalid("At least one page must not be skipped")
        buffer_frac = config[CONF_BUFFER_SIZE]
        if CORE.is_esp32 and buffer_frac > 0.5 and PSRAM_DOMAIN not in global_config:
            df.LOGGER.warning("buffer_size: may need to be reduced without PSRAM")

    if len(config_list) != 1:
        multi_conf_validate(config_list)

    for w in get_focused_widgets():
        path = global_config.get_path_for_id(w)
        widget_conf = global_config.get_config_for_path(path[:-1])
        if df.CONF_ADJUSTABLE in widget_conf and not widget_conf[df.CONF_ADJUSTABLE]:
            raise cv.Invalid(
                "A non adjustable arc may not be focused",
                path,
            )
    for w in get_refreshed_widgets():
        path = global_config.get_path_for_id(w)
        widget_conf = global_config.get_config_for_path(path[:-1])
        if not any(isinstance(v, (Lambda, dict)) for v in widget_conf.values()):
            raise cv.Invalid(
                f"Widget '{w}' does not have any dynamic properties to refresh",
            )
    # Do per-widget type final validation for update actions
    for widget_type, update_configs in df.get_updated_widgets().items():
        for conf in update_configs:
            for id_conf in conf.get(CONF_ID, ()):
                name = id_conf[CONF_ID]
                path = global_config.get_path_for_id(name)
                widget_conf = global_config.get_config_for_path(path[:-1])
                widget_type.final_validate(name, conf, widget_conf, path[1:])


async def to_code(configs):
    config_0 = configs[0]
    # Global configuration
    if CORE.is_esp32:
        # Skip compiling lvgl examples
        add_idf_sdkconfig_option("CONFIG_LV_BUILD_EXAMPLES", False)
        add_idf_sdkconfig_option("CONFIG_LV_BUILD_DEMOS", False)
        if get_esp32_variant() == VARIANT_ESP32P4:
            add_idf_sdkconfig_option("CONFIG_LV_DRAW_BUF_ALIGN", 64)
            # disable use of PPA for fills until upstream bugs fixed
            df.add_define("LV_USE_PPA", "0")
            df.add_define("LV_DRAW_BUF_ALIGN", "64")
        else:
            df.add_define("LV_DRAW_BUF_ALIGN", "32")
        add_idf_component(name="lvgl/lvgl", ref=LVGL_VERSION)
    else:
        df.add_define("LV_DRAW_BUF_ALIGN", "1")
        cg.add_library("lvgl/lvgl", LVGL_VERSION)
    df.add_define("LV_DRAW_BUF_STRIDE_ALIGN", "1")
    df.add_define("LV_USE_DRAW_SW", "1")
    df.add_define("LV_USE_STDLIB_SPRINTF", "LV_STDLIB_CLIB")
    df.add_define("LV_USE_STDLIB_STRING", "LV_STDLIB_CLIB")
    df.add_define("LV_USE_STDLIB_MALLOC", "LV_STDLIB_CUSTOM")
    df.add_define("LV_DEF_REFR_PERIOD", "16")
    cg.add_define("USE_LVGL")
    # suppress default enabling of extra widgets
    # cg.add_define("LV_KCONFIG_PRESENT")
    # Always enable - lots of things use it.
    df.add_define("LV_DRAW_SW_COMPLEX", "1")

    df.add_define(
        "LV_LOG_LEVEL",
        f"LV_LOG_LEVEL_{df.LV_LOG_LEVELS[config_0[CONF_LOG_LEVEL]]}",
    )
    df.add_define("LV_USE_LOG", "1")
    cg.add_define(
        "LVGL_LOG_LEVEL",
        cg.RawExpression(f"ESPHOME_LOG_LEVEL_{config_0[CONF_LOG_LEVEL]}"),
    )
    df.add_define("LV_COLOR_DEPTH", config_0[CONF_COLOR_DEPTH])
    for font in df.get_lv_fonts_used():
        df.add_define(f"LV_FONT_{font.upper()}")

    if config_0[CONF_COLOR_DEPTH] == 16:
        df.add_define(
            "LV_COLOR_16_SWAP",
            "1" if config_0[CONF_BYTE_ORDER] == "big_endian" else "0",
        )
    df.add_define(
        "LV_COLOR_CHROMA_KEY",
        await lvalid.lv_color.process(config_0[df.CONF_TRANSPARENCY_KEY]),
    )
    cg.add_build_flag("-Isrc")

    cg.add_global(lvgl_ns.using)
    for font in df.get_esphome_fonts_used():
        await cg.get_variable(font)
    default_font = config_0[df.CONF_DEFAULT_FONT]
    if not lvalid.is_lv_font(default_font):
        df.add_define(
            "LV_FONT_CUSTOM_DECLARE", f"LV_FONT_DECLARE(*{df.DEFAULT_ESPHOME_FONT})"
        )
        globfont_id = ID(
            df.DEFAULT_ESPHOME_FONT,
            True,
            type=lv_font_t.operator("ptr").operator("const"),
        )
        # static=False because LV_FONT_CUSTOM_DECLARE creates an extern declaration
        cg.new_variable(
            globfont_id,
            MockObj(await lvalid.lv_font.process(default_font), "->").get_lv_font(),
            static=False,
        )
        df.add_define("LV_FONT_DEFAULT", df.DEFAULT_ESPHOME_FONT)
    else:
        df.add_define("LV_FONT_DEFAULT", await lvalid.lv_font.process(default_font))
    cg.add(lvgl_static.esphome_lvgl_init())
    default_group = get_default_group(config_0)

    for config in configs:
        frac = config[CONF_BUFFER_SIZE]
        if frac >= 0.75:
            frac = 1
        elif frac >= 0.375:
            frac = 2
        elif frac > 0.19:
            frac = 4
        elif frac != 0:
            frac = 8
        displays = [
            await cg.get_variable(display) for display in config[df.CONF_DISPLAYS]
        ]
        rotation_type = RotationType.ROTATION_UNUSED
        # options will have CONF_ROTATION true if rotation is changed in an automation.
        if CONF_ROTATION in config or df.get_options().get(CONF_ROTATION) is True:
            if all(
                get_display_metadata(disp).has_hardware_rotation for disp in displays
            ):
                rotation_type = RotationType.ROTATION_HARDWARE
                df.LOGGER.info("LVGL will use hardware rotation via display driver")
            else:
                rotation_type = RotationType.ROTATION_SOFTWARE
                if CORE.is_esp32 and get_esp32_variant() == VARIANT_ESP32P4:
                    df.LOGGER.info("LVGL will use software rotation (PPA accelerated)")
                else:
                    df.LOGGER.info("LVGL will use software rotation")
        lv_component = cg.new_Pvariable(
            config[CONF_ID],
            displays,
            frac,
            config[df.CONF_FULL_REFRESH],
            config[CONF_DRAW_ROUNDING],
            config[df.CONF_RESUME_ON_INPUT],
            config[df.CONF_UPDATE_WHEN_DISPLAY_IDLE],
            rotation_type,
        )
        await cg.register_component(lv_component, config)
        if rotation := config.get(CONF_ROTATION):
            cg.add(lv_component.set_rotation(rotation))
        if paused := config[df.CONF_PAUSED]:
            cg.add(lv_component.set_paused(paused, False))
        if refr_time := config.get(df.CONF_REFRESH_INTERVAL):
            cg.add(lv_component.set_refresh_interval(refr_time.total_milliseconds))
        Widget.create(config[CONF_ID], lv_component, LvScrActType(), config)

        lv_scr_act = get_screen_active(lv_component)
        async with LvContext():
            cg.add(lv_component.set_big_endian(config[CONF_BYTE_ORDER] == "big_endian"))
            await touchscreens_to_code(lv_component, config)
            await encoders_to_code(lv_component, config, default_group)
            await keypads_to_code(lv_component, config, default_group)
            await theme_to_code(config)
            await gradients_to_code(config)
            await styles_to_code(config)
            await set_obj_properties(lv_scr_act, config)
            await add_widgets(lv_scr_act, config)
            await add_pages(lv_component, config)
            await layers_to_code(lv_component, config)
            await lvgl_update(lv_component, config)
            await msgboxes_to_code(lv_component, config)
            await animations_to_code(config.get(CONF_ANIMATIONS, []))

    # Mark all widgets as completed so awaiters of ``wait_for_widgets`` proceed.
    set_widgets_completed(True)
    async with LvContext():
        await generate_triggers()
        await generate_align_tos(configs[0])
        for config in configs:
            lv_component = await cg.get_variable(config[CONF_ID])
            await add_animation_triggers(config.get(CONF_ANIMATIONS, []))
            await generate_page_triggers(config)
            await initial_focus_to_code(config)
            for conf in config.get(CONF_ON_IDLE, ()):
                templ = await cg.templatable(conf[CONF_TIMEOUT], [], cg.uint32)
                idle_trigger = cg.new_Pvariable(
                    conf[CONF_TRIGGER_ID], lv_component, templ
                )
                await build_automation(idle_trigger, [], conf)
            for trigger_name in SIMPLE_TRIGGERS:
                if conf := config.get(trigger_name):
                    trigger_var = cg.new_Pvariable(conf[CONF_TRIGGER_ID])
                    await build_automation(trigger_var, [], conf)
                    cg.add(
                        getattr(
                            lv_component,
                            f"set_{trigger_name.removeprefix('on_')}_trigger",
                        )(trigger_var)
                    )

    # This must be done after all widgets are created
    styles_used = df.get_styles_used()
    if any(BASE_PROPS.get(x) is lvalid.lv_image for x in styles_used):
        add_lv_use(CONF_IMAGE)
    for use in df.get_lv_uses():
        df.add_define(f"LV_USE_{use.upper()}")
        cg.add_define(f"USE_LVGL_{use.upper()}")

    if {
        "transform_rotation",
        "transform_scale",
        "transform_scale_x",
        "transform_scale_y",
    } & styles_used:
        df.add_define("LV_COLOR_SCREEN_TRANSP", "1")

    if configs[0].get(df.CONF_THEME, {}).get(df.CONF_DARK_MODE):
        df.add_define("LV_THEME_DEFAULT_DARK", "1")

    # Currently always need RGB565 for the display buffer, and ARGB8888 is used for layer blending
    lv_image_formats = {"RGB565", "ARGB8888"}
    if {
        "drop_shadow_color",
        "drop_shadow_offset_x",
        "drop_shadow_offset_y",
        "drop_shadow_opa",
        "drop_shadow_quality",
        "drop_shadow_radius",
    } & styles_used:
        lv_image_formats.add("A8")

    for image_id in get_lv_images_used():
        await cg.get_variable(image_id)
        metadata = get_image_metadata(image_id.id)
        image_type = IMAGE_TYPE[metadata.image_type]
        transparent = metadata.transparency != CONF_OPAQUE
        if image_type == ImageBinary:
            lv_image_formats.add("I1")
        if image_type == ImageGrayscale:
            lv_image_formats.add("A8")
        if image_type == ImageRGB565:
            lv_image_formats.add("RGB565A8" if transparent else "RGB565")
        if image_type == ImageRGB:
            lv_image_formats.add("ARGB8888" if transparent else "RGB8888")
    if df.is_defined("LV_GRADIENT_MAX_STOPS"):
        lv_image_formats.add("RGB888")
    for fmt in lv_image_formats:
        df.add_define(f"LV_DRAW_SW_SUPPORT_{fmt}", "1")

    lv_conf_h_file = CORE.relative_src_path(LV_CONF_FILENAME)
    if write_file_if_changed(lv_conf_h_file, generate_lv_conf_h()):
        clean_build(clear_pio_cache=False)
    cg.add_build_flag("-DLV_CONF_H=1")
    # handle windows paths in a way that doesn't break the generated C++
    lv_conf_h_path = Path(lv_conf_h_file).as_posix()
    cg.add_build_flag(f'-DLV_CONF_PATH=\\"{lv_conf_h_path}\\"')
    cg.add_build_flag("-DLV_KCONFIG_IGNORE")

    for prop in df.get_remapped_uses():
        df.LOGGER.warning(
            "Property '%s' is deprecated, use '%s' instead", prop, STYLE_REMAP[prop]
        )
    for warning in df.get_warnings():
        df.LOGGER.warning(warning)


def display_schema(config):
    value = cv.ensure_list(cv.use_id(Display))(config)
    value = value or [cv.use_id(Display)(config)]
    if len(set(value)) != len(value):
        raise cv.Invalid("Display IDs must be unique")
    return value


def add_hello_world(config):
    if df.CONF_WIDGETS not in config and CONF_PAGES not in config:
        df.LOGGER.info(
            "No pages or widgets configured, creating default hello_world page"
        )
        hello_world_path = Path(__file__).parent / HELLO_WORLD_FILE
        config[df.CONF_WIDGETS] = any_widget_schema()(load_yaml(hello_world_path))
    return config


@functools.cache
def _build_theme_schema(
    widget_types: tuple[tuple[str, widgets.WidgetType], ...],
) -> cv.Schema:
    # The theme schema is value-independent: it depends only on the set of
    # registered widget types. Key the cache on a snapshot of WIDGET_TYPES so
    # that an external component registering a new widget after the first
    # validation (legal per any_widget_schema's lazy-evaluation contract)
    # produces a fresh tuple, a cache miss, and a rebuilt schema -- the cache
    # self-heals instead of stale-rejecting valid themes. See obj_dict() in
    # schemas.py for why chained .extend() is avoided here.
    return cv.Schema(
        {
            cv.Optional(df.CONF_DARK_MODE, default=False): cv.boolean,
            **{
                cv.Optional(name): cv.Schema(
                    {**obj_dict(w), **FULL_STYLE_SCHEMA.schema}
                )
                for name, w in widget_types
            },
        }
    )


def _theme_schema(value: dict) -> dict:
    return _build_theme_schema(tuple(WIDGET_TYPES.items()))(value)


FINAL_VALIDATE_SCHEMA = final_validation

# The options accepted at the top level of an `lvgl:` block, on top of the base
# object schema that `container_schema(obj_spec, ...)` supplies. Held in a
# module-level name (rather than inline) so the schema-extractor wrapper on
# CONFIG_SCHEMA below can hand the language-schema dumper the same composed
# schema the runtime validates against.
LVGL_TOP_LEVEL_SCHEMA = (
    cv.polling_component_schema("1s")
    .extend(
        {
            **{
                cv.Optional(event): validate_automation(
                    {
                        cv.GenerateID(CONF_TRIGGER_ID): cv.declare_id(
                            Trigger.template(lv_obj_t_ptr, lv_event_t_ptr)
                        ),
                    }
                )
                for event in df.LV_SCREEN_EVENT_TRIGGERS + df.LV_DISPLAY_EVENT_TRIGGERS
            },
            cv.GenerateID(CONF_ID): cv.declare_id(LvglComponent),
            cv.GenerateID(CONF_ALIGN_TO_LAMBDA_ID): cv.declare_id(lv_lambda_t),
            cv.GenerateID(df.CONF_DISPLAYS): display_schema,
            cv.Optional(CONF_COLOR_DEPTH, default=16): cv.one_of(16),
            cv.Optional(df.CONF_DEFAULT_FONT, default="montserrat_14"): lvalid.lv_font,
            cv.Optional(df.CONF_FULL_REFRESH, default=False): cv.boolean,
            cv.Optional(df.CONF_UPDATE_WHEN_DISPLAY_IDLE, default=False): cv.boolean,
            cv.Optional(df.CONF_REFRESH_INTERVAL): cv.positive_time_period_milliseconds,
            cv.Optional(CONF_DRAW_ROUNDING, default=2): cv.positive_int,
            cv.Optional(CONF_BUFFER_SIZE, default=0): cv.percentage,
            cv.Optional(CONF_ROTATION): validate_rotation,
            cv.Optional(CONF_LOG_LEVEL, default="WARN"): cv.one_of(
                *df.LV_LOG_LEVELS, upper=True
            ),
            cv.Optional(CONF_BYTE_ORDER): cv.one_of(
                "big_endian", "little_endian", lower=True
            ),
            cv.Optional(df.CONF_STYLE_DEFINITIONS): cv.ensure_list(
                cv.Schema({cv.Required(CONF_ID): cv.declare_id(lv_style_t)}).extend(
                    FULL_STYLE_SCHEMA
                )
            ),
            cv.Optional(CONF_ON_IDLE): validate_automation(
                {
                    cv.GenerateID(CONF_TRIGGER_ID): cv.declare_id(IdleTrigger),
                    cv.Required(CONF_TIMEOUT): cv.templatable(
                        cv.positive_time_period_milliseconds
                    ),
                }
            ),
            cv.Optional(CONF_PAGES): cv.ensure_list(container_schema(page_spec)),
            **{
                cv.Optional(x): validate_automation(
                    {
                        cv.GenerateID(CONF_TRIGGER_ID): cv.declare_id(PlainTrigger),
                    },
                    single=True,
                )
                for x in SIMPLE_TRIGGERS
            },
            cv.Optional(df.CONF_MSGBOXES): cv.ensure_list(MSGBOX_SCHEMA),
            cv.Optional(df.CONF_ANIMATIONS): cv.ensure_list(ANIMATION_SCHEMA),
            cv.Optional(df.CONF_PAGE_WRAP, default=True): lv_bool,
            cv.Optional(df.CONF_TOP_LAYER): container_schema(obj_spec),
            cv.Optional(df.CONF_BOTTOM_LAYER): container_schema(obj_spec),
            cv.Optional(df.CONF_TRANSPARENCY_KEY, default=0x000400): lvalid.lv_color,
            cv.Optional(df.CONF_THEME): _theme_schema,
            cv.Optional(df.CONF_GRADIENTS): GRADIENT_SCHEMA,
            cv.Optional(df.CONF_TOUCHSCREENS, default=None): touchscreen_schema,
            cv.Optional(df.CONF_ENCODERS, default=None): ENCODERS_CONFIG,
            cv.Optional(df.CONF_KEYPADS, default=None): KEYPADS_CONFIG,
            cv.GenerateID(df.CONF_DEFAULT_GROUP): cv.declare_id(lv_group_t),
            cv.Optional(df.CONF_RESUME_ON_INPUT, default=True): cv.boolean,
            cv.Optional(df.CONF_PAUSED, default=False): cv.boolean,
        }
    )
    .extend(DISP_BG_SCHEMA)
)


LVGL_SCHEMA = cv.All(
    container_schema(obj_spec, LVGL_TOP_LEVEL_SCHEMA),
    cv.has_at_most_one_key(CONF_PAGES, df.CONF_LAYOUT),
    add_hello_world,
)


@schema_extractor("schema")
def lvgl_config_schema(config):
    """
    Can't use cv.ensure_list here because it converts an empty config to an empty list,
    rather than a default config.
    """
    if config is SCHEMA_EXTRACT:
        # CONFIG_SCHEMA is this callable wrapping `cv.All` over a container_schema
        # closure, so the language-schema dumper can't see the top-level `lvgl:`
        # fields (it would emit an empty schema). Hand it the same composed
        # obj + top-level schema the runtime validates against, plus the
        # `widgets:` key (added per-value by append_layout_schema at runtime, so
        # otherwise invisible to the dumper). Validation of real configs (the
        # branches below) is unchanged.
        return container_schema_value(obj_spec, LVGL_TOP_LEVEL_SCHEMA).extend(
            {cv.Optional(df.CONF_WIDGETS): any_widget_schema()}
        )
    if not config or isinstance(config, dict):
        return [LVGL_SCHEMA(config)]
    return cv.Schema([LVGL_SCHEMA])(config)


CONFIG_SCHEMA = lvgl_config_schema



from esphome import automation, core
import esphome.codegen as cg
from esphome.components import audio, speaker
import esphome.config_validation as cv
from esphome.const import (
    CONF_BITS_PER_SAMPLE,
    CONF_ID,
    CONF_NUM_CHANNELS,
    CONF_OUTPUT_SPEAKER,
    CONF_SAMPLE_RATE,
)
from esphome.core import ID
from esphome.cpp_generator import MockObj
from esphome.types import ConfigType, TemplateArgsType

CODEOWNERS = ["@kahrendt"]

CONF_OUTPUT_SPEAKERS = "output_speakers"
CONF_TARGET_SPEAKER = "target_speaker"

router_ns = cg.esphome_ns.namespace("router")
Router = router_ns.class_("Router", cg.Component, speaker.Speaker)
SwitchOutputAction = router_ns.class_("SwitchOutputAction", automation.Action)

SpeakerPtr = speaker.Speaker.operator("ptr")


def _set_stream_limits(config: ConfigType) -> ConfigType:
    # Lock the router's stream limits to the user-declared format. Limits are set
    # at CONFIG_SCHEMA time so they're visible to other components' FINAL_VALIDATE
    # (which has no guaranteed ordering vs. ours).
    audio.set_stream_limits(
        min_bits_per_sample=config[CONF_BITS_PER_SAMPLE],
        max_bits_per_sample=config[CONF_BITS_PER_SAMPLE],
        min_channels=config[CONF_NUM_CHANNELS],
        max_channels=config[CONF_NUM_CHANNELS],
        min_sample_rate=config[CONF_SAMPLE_RATE],
        max_sample_rate=config[CONF_SAMPLE_RATE],
    )(config)
    return config


CONFIG_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(Router),
            cv.Required(CONF_OUTPUT_SPEAKERS): cv.All(
                cv.ensure_list(cv.use_id(speaker.Speaker)),
                cv.Length(min=2, max=8),
            ),
            # All outputs must agree on a single format so the producer can keep
            # streaming through a switch without reconfiguring. These are required
            # rather than inherited because downstream components (e.g. mixer)
            # read them from the router's declaration during FINAL_VALIDATE,
            # which can't depend on our FINAL_VALIDATE running first.
            cv.Required(CONF_BITS_PER_SAMPLE): cv.int_range(8, 32),
            cv.Required(CONF_NUM_CHANNELS): cv.int_range(1, 2),
            cv.Required(CONF_SAMPLE_RATE): cv.int_range(8000, 96000),
        }
    ).extend(cv.COMPONENT_SCHEMA),
    cv.only_on_esp32,
    _set_stream_limits,
)


def _final_validate(config: ConfigType) -> ConfigType:
    # Validate every configured output speaker can accept the router's format.
    # Switching to an output that can't reproduce the format the producer is
    # already sending would otherwise fail silently at runtime.
    for spk_id in config[CONF_OUTPUT_SPEAKERS]:
        proxy = {**config, CONF_OUTPUT_SPEAKER: spk_id}
        audio.final_validate_audio_schema(
            "router",
            audio_device=CONF_OUTPUT_SPEAKER,
            bits_per_sample=config[CONF_BITS_PER_SAMPLE],
            channels=config[CONF_NUM_CHANNELS],
            sample_rate=config[CONF_SAMPLE_RATE],
        )(proxy)
    return config


FINAL_VALIDATE_SCHEMA = _final_validate


async def to_code(config: ConfigType) -> None:
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)

    # The first configured output is the default active output on boot.
    speakers = config[CONF_OUTPUT_SPEAKERS]
    cg.add(var.set_output_count(len(speakers)))
    for spk_id in speakers:
        spk = await cg.get_variable(spk_id)
        cg.add(var.add_output(spk))


@automation.register_action(
    "router.speaker.switch_output",
    SwitchOutputAction,
    cv.Schema(
        {
            cv.GenerateID(CONF_ID): cv.use_id(Router),
            cv.Required(CONF_TARGET_SPEAKER): cv.templatable(
                cv.use_id(speaker.Speaker)
            ),
        }
    ),
    synchronous=True,
)
async def switch_output_to_code(
    config: ConfigType,
    action_id: ID,
    template_arg: cg.TemplateArguments,
    args: TemplateArgsType,
) -> MockObj:
    parent = await cg.get_variable(config[CONF_ID])
    var = cg.new_Pvariable(action_id, template_arg, parent)
    target = config[CONF_TARGET_SPEAKER]
    if not isinstance(target, core.Lambda):
        target = await cg.get_variable(target)
    template_ = await cg.templatable(target, args, SpeakerPtr)
    cg.add(var.set_target(template_))
    return var



import esphome.codegen as cg
from esphome.components import text_sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_ID,
    CONF_MAC_ADDRESS,
    CONF_VERSION,
    ENTITY_CATEGORY_DIAGNOSTIC,
    ICON_BLUETOOTH,
    ICON_CHIP,
)

from . import CONF_LD2410_ID, LD2410Component

DEPENDENCIES = ["ld2410"]

CONFIG_SCHEMA = {
    cv.GenerateID(CONF_ID): cv.declare_id(cg.EntityBase),
    cv.GenerateID(CONF_LD2410_ID): cv.use_id(LD2410Component),
    cv.Optional(CONF_VERSION): text_sensor.text_sensor_schema(
        entity_category=ENTITY_CATEGORY_DIAGNOSTIC, icon=ICON_CHIP
    ),
    cv.Optional(CONF_MAC_ADDRESS): text_sensor.text_sensor_schema(
        entity_category=ENTITY_CATEGORY_DIAGNOSTIC, icon=ICON_BLUETOOTH
    ),
}


async def to_code(config):
    ld2410_component = await cg.get_variable(config[CONF_LD2410_ID])
    if version_config := config.get(CONF_VERSION):
        sens = await text_sensor.new_text_sensor(version_config)
        cg.add(ld2410_component.set_version_text_sensor(sens))
    if mac_address_config := config.get(CONF_MAC_ADDRESS):
        sens = await text_sensor.new_text_sensor(mac_address_config)
        cg.add(ld2410_component.set_mac_text_sensor(sens))



"""Toolchain-agnostic PlatformIO library converter.

Resolves a batch of PlatformIO/Arduino library specs (added via
``cg.add_library(...)``) into local, build-ready directories: it fetches each
library (registry/git/url), parses its ``library.json`` / ``library.properties``
manifest, resolves the whole dependency graph to a single version per name, and
caches the result under ``<data_dir>/pio_components``.

The toolchain-specific part — turning a resolved library into build files
(ESP-IDF ``idf_component_register`` CMakeLists, or a Zephyr module) — is supplied
by a :class:`LibraryBackend`. This module owns everything that is the same
regardless of which toolchain consumes the result.
"""

from collections import deque
from collections.abc import Callable
from dataclasses import dataclass, field
import glob
import hashlib
import itertools
import json
import logging
import os
from pathlib import Path
import re
import tempfile
from typing import Any
from urllib.parse import urlsplit, urlunsplit

from esphome import git
from esphome.core import CORE, Library
from esphome.framework_helpers import archive_extract_all, download_from_mirrors, rmdir

_LOGGER = logging.getLogger(__name__)

PathType = str | os.PathLike

#
# Constants from platformio
#

FILTER_REGEX = re.compile(r"([+-])<([^>]+)>")
DEFAULT_BUILD_SRC_FILTER = (
    "+<*> -<.git/> -<.svn/> -<example/> -<examples/> -<test/> -<tests/>"
)
DEFAULT_BUILD_SRC_DIRS = "src"
DEFAULT_BUILD_INCLUDE_DIR = "include"
DEFAULT_BUILD_FLAGS = []
SRC_FILE_EXTENSIONS = [
    ".c",
    ".cpp",
    ".cc",
    ".cxx",
    ".c++",
    ".S",
    ".spp",
    ".SPP",
    ".sx",
    ".s",
    ".asm",
    ".ASM",
]

DOMAIN = "pio_components"

ESPHOME_DATA_KEY = "ESPHOME"
ESPHOME_DATA_EXTRA_CMAKE_KEY = "EXTRA_CMAKE"


class Source:
    def download(
        self, dir_suffix: str, force: bool = False, salt: str = "", namespace: str = ""
    ) -> Path:
        raise NotImplementedError


class URLSource(Source):
    def __init__(self, url: str):
        self.url = url

    def download(
        self, dir_suffix: str, force: bool = False, salt: str = "", namespace: str = ""
    ) -> Path:
        # Namespace the cache per backend (e.g. pio_components/idf, .../zephyr) so
        # the build files each backend writes into the library dir can't collide.
        base_dir = Path(CORE.data_dir) / DOMAIN
        if namespace:
            base_dir = base_dir / namespace
        h = hashlib.new("sha256")
        h.update(self.url.encode())
        if salt:
            h.update(salt.encode())
        path = base_dir / h.hexdigest()[:8] / dir_suffix
        # Marker file written last to signal a complete extraction. Using a
        # marker (instead of just `path.is_dir()`) means an interrupted
        # extraction is correctly detected and re-run on the next invocation,
        # and lets us extract directly into ``path`` — avoiding a
        # post-extraction rename that races with antivirus on Windows.
        extracted_marker = path / ".esphome_extracted"
        if not extracted_marker.is_file() or force:
            rmdir(path, msg=f"Clean up library directory {path}")

            # Download in temporary file
            with tempfile.NamedTemporaryFile() as tmp:
                _LOGGER.info("Downloading %s ...", self.url)
                _LOGGER.debug("Location: %s", path)

                download_from_mirrors([self.url], {}, tmp.file)

                _LOGGER.debug("Extracting archive to %s ...", path)
                archive_extract_all(tmp.file, path)
                extracted_marker.touch()
        return path

    def __str__(self):
        return self.url


class GitSource(Source):
    def __init__(self, url: str, ref: str | None):
        self.url = url
        self.ref = ref

    def download(
        self, dir_suffix: str, force: bool = False, salt: str = "", namespace: str = ""
    ) -> Path:
        domain = DOMAIN
        if namespace:
            domain = f"{domain}/{namespace}"
        if salt:
            domain = f"{domain}/{salt}"
        path, _ = git.clone_or_update(
            url=self.url,
            ref=self.ref,
            refresh=git.NEVER_REFRESH if not force else None,
            domain=domain,
            init_submodules=True,
            subpath=Path(dir_suffix),
        )
        return path

    def __str__(self):
        return f"{self.url}#{self.ref}" if self.ref else self.url


class InvalidLibrary(Exception):
    pass


class ConvertedLibrary:
    """A resolved PlatformIO library plus its parsed manifest and on-disk path.

    Toolchain-neutral: ESP-IDF treats it as a component, Zephyr as a module. The
    backend reads ``name``/``version``/``data``/``dependencies``/``path`` to emit
    its build files.
    """

    def __init__(self, name: str, version: str, source: Source | None):
        self.name = name
        self.version = version
        self.source = source
        self.data = {}
        self.dependencies: list[ConvertedLibrary] = []
        self._path: Path | None = None

    def __str__(self):
        return f"{self.name}@{self.version}={self.source}"

    @property
    def path(self) -> Path:
        if self._path is None:
            raise RuntimeError(f"path not set for library {self}")
        return self._path

    @path.setter
    def path(self, value: Path) -> None:
        self._path = value

    def get_sanitized_name(self):
        return re.sub(r"[^a-zA-Z0-9_.\-/]", "_", self.name)

    def get_require_name(self):
        return self.get_sanitized_name().replace("/", "__")

    def download(self, force: bool = False, salt: str = "", namespace: str = ""):
        """Fetch the library into the shared cache and record its ``path``.

        The cache directory is named after the sanitized library name; backends
        rely on that name to identify the unit they build (e.g. ESP-IDF uses the
        directory name as the component name, replacing ``/`` with ``__`` via
        ``get_require_name``). ``namespace`` keeps each backend's cache separate.
        """
        self.path = self.source.download(
            self.get_sanitized_name(), force=force, salt=salt, namespace=namespace
        )


@dataclass
class LibraryBackend:
    """Toolchain hooks for :func:`convert_libraries`.

    ``platform``/``framework`` drive the manifest compatibility check.
    ``emit`` writes the toolchain-specific build files into a resolved library's
    ``path`` (e.g. the ESP-IDF ``CMakeLists.txt`` + ``idf_component.yml``, or a
    Zephyr ``module.yml`` + ``CMakeLists.txt``).
    ``cache_key`` namespaces the download cache (``pio_components/<cache_key>/``)
    so the differing build files two backends emit into a library dir never
    collide when the same config dir hosts both an ESP-IDF and a Zephyr build.
    """

    platform: str | None
    framework: str
    emit: Callable[["ConvertedLibrary"], None]
    cache_key: str


def ensure_list[T](obj: T | list[T]) -> list[T]:
    """
    Convert an object to a list if it isn't already a list.

    Args:
        obj: Object that may or may not already be a list.

    Returns:
        list[T]: The original list if ``obj`` is a list, otherwise a single-item
        list containing ``obj``.
    """
    return [obj] if not isinstance(obj, list) else obj


def _owner_pkgname_to_name(owner: str | None, pkgname: str) -> str:
    """
    Convert owner and package name to a standardized component name.

    This function combines owner and package name with a forward slash when
    both are provided, otherwise returns just the package name.

    Args:
        owner: The owner/username of the package (can be None)
        pkgname: The name of the package

    Returns:
        str: The standardized component name in "owner/pkgname" format or just "pkgname"
    """
    return f"{owner}/{pkgname}" if owner else pkgname


def collect_filtered_files(src_dir: PathType, src_filters: list[str]) -> list[str]:
    """
    Recursively match files in a directory according to include/exclude patterns.

    This function processes a list of filter strings that indicate which files
    to include or exclude. Each filter is parsed into patterns with a sign:
    '+' for inclusion and '-' for exclusion. Directory patterns ending with '/'
    are normalized to include all their contents recursively.

    Args:
        src_dir (PathType): Root directory to search within.
        src_filters (list[str]): List of filter strings, which may contain multiple
            patterns. Each pattern can start with '+' or '-' to indicate inclusion
            or exclusion.

    Returns:
        list[str]: List of matched file paths as strings. Only files (not directories)
        are returned, even if a directory matches a pattern.
    """
    matches = list(
        itertools.chain.from_iterable(
            FILTER_REGEX.findall(src_filter) for src_filter in src_filters
        )
    )

    selected = set()

    for sign, pattern in matches:
        pattern = pattern.strip()

        if pattern.endswith("/"):
            pattern = pattern.rstrip("/") + "/**"

        # glob.escape has no pathlib equivalent and the matcher works on raw
        # path strings, so PTH118/PTH207 don't apply here.
        full_pattern = os.path.join(glob.escape(str(src_dir)), pattern)  # noqa: PTH118

        matched = []
        for item in glob.glob(full_pattern, recursive=True):  # noqa: PTH207
            if not Path(item).is_dir():
                matched.append(item)
            else:
                # PlatformIO quirk: a directory matched with "*" should include all its
                # nested files and subdirectories, not just the directory itself.
                for root, _, files in os.walk(item):
                    matched.extend([str(Path(root) / f) for f in files])

        # glob keeps the pattern's literal separators for non-wildcard path
        # components, so on Windows the same file can surface with different
        # separators depending on where the wildcards sit; normalize so the
        # include/exclude set operations below compare equal paths.
        matched = [os.path.normpath(m) for m in matched]

        # FILTER_REGEX only ever captures "+" or "-", so the else is the "-" case.
        if sign == "+":
            selected.update(matched)
        else:
            selected.difference_update(matched)

    return [r for r in selected if Path(r).is_file()]


def split_list_by_condition(
    items: list[str], match_fn: Callable[[str], str | None]
) -> tuple[list[str], list[str]]:
    """
    Splits a list into two lists based on a matching function.

    Args:
        items: List of items to split.
        match_fn: Function that returns a value for items that should go into the "matched" list.

    Returns:
        A tuple (matched, non_matched)
    """
    matched = []
    non_matched = []
    for item in items:
        result = match_fn(item)
        if result:
            matched.append(result)
        else:
            non_matched.append(item)
    return matched, non_matched


def check_library_data(data: dict, platform: str | None, framework: str):
    """
    Check whether a library manifest is compatible with the target toolchain.

    A platform mismatch (e.g. an AVR-only library on ESP32) raises
    ``InvalidLibrary`` so the caller skips the library. A framework mismatch only
    logs a warning — PIO manifests often understate the frameworks they actually
    compile under, and there's no opt-out at this layer, so we include the library
    anyway.

    Args:
        data: PIO library manifest dict being processed.
        platform: The PlatformIO platform token the build targets (e.g.
            ``espressif32``). ``None`` skips the platform check entirely — useful
            for targets (e.g. Zephyr) where PIO manifests rarely declare the
            platform yet portable libraries still build.
        framework: The active framework name (e.g. ``espidf``, ``arduino``,
            ``zephyr``) the manifest is expected to declare.

    Raises:
        InvalidLibrary: If the library does not support the target platform.
    """
    platforms = data.get("platforms", "*")
    if isinstance(platforms, str):
        platforms = [a.strip() for a in platforms.split(",")]
    platforms = ensure_list(platforms)

    # Check if library supports the target platform
    valid_platforms = platform is None or "*" in platforms or platform in platforms

    if not valid_platforms:
        raise InvalidLibrary(f"Unsupported library platforms: {platforms}")

    frameworks = data.get("frameworks", "*")
    if isinstance(frameworks, str):
        frameworks = [a.strip() for a in frameworks.split(",")]
    frameworks = ensure_list(frameworks)

    # Check if library declares the active framework. PIO library manifests
    # often list only "arduino" even when the library actually compiles fine
    # under the target framework, and there's no way to opt out of the check at
    # this layer. Warn instead of failing so the user isn't forced to fork the
    # library to fix the manifest.
    valid_framework = "*" in frameworks or framework in frameworks

    if not valid_framework:
        _LOGGER.warning(
            "Library %s declares frameworks %s that do not include '%s'; including anyway",
            data.get("name", "<unknown>"),
            frameworks,
            framework,
        )


def _parse_library_json(library_json_path: PathType):
    """
    Load and parse a JSON file describing a library.

    Args:
        library_json_path (PathType): Path to the JSON file.

    Returns:
        dict: Parsed JSON content as a Python dictionary.
    """
    with Path(library_json_path).open(encoding="utf8") as fp:
        return json.load(fp)


def _parse_library_properties(library_properties_path: PathType):
    """
    Parse a key-value platformio .properties style file into a dictionary.

    Args:
        library_properties_path (PathType): Path to the properties file.

    Returns:
        dict[str, str]: Mapping of parsed property keys to values.
    """
    with Path(library_properties_path).open(encoding="utf8") as fp:
        data = {}
        for line in fp.read().splitlines():
            line = line.strip()
            if not line or "=" not in line:
                continue
            # skip comments
            if line.startswith("#"):
                continue
            key, value = line.split("=", 1)
            if not value.strip():
                continue
            data[key.strip()] = value.strip()
        return data


def _make_registry_client() -> Any:
    """Create a minimal PlatformIO registry client with no system filtering.

    ``is_system_compatible`` is forced True so version selection is driven purely
    by the requested version requirements -- target compatibility is handled
    elsewhere, not by the PlatformIO registry.
    """
    from platformio.package.manager._registry import PackageManagerRegistryMixin

    class _Registry(PackageManagerRegistryMixin):
        def __init__(self) -> None:
            self._registry_client = None
            self.pkg_type = "library"

        @staticmethod
        def is_system_compatible(value: Any, custom_system: Any = None) -> bool:
            return True

    return _Registry()


def _resolve_registry_version(
    owner: str | None, pkgname: str, requirements: set[str]
) -> tuple[str, str, str, str]:
    """Resolve a registry package to the single highest version satisfying ALL
    the given requirements; return ``(owner, name, version, download_url)``.

    Intersecting every requirement (rather than resolving each consumer in
    isolation) makes the result independent of processing order and guarantees
    no stated constraint is violated -- e.g. ``esphome/libsodium`` requested as
    both ``==1.10021.0`` and ``^1.10018.1`` resolves to ``1.10021.0``.
    """
    from platformio.package.meta import PackageSpec

    registry = _make_registry_client()
    package = registry.fetch_registry_package(PackageSpec(owner=owner, name=pkgname))
    owner = package["owner"]["username"]
    name = package["name"]

    # Chaining the per-requirement filter intersects all constraints.
    versions = package.get("versions") or []
    for requirement in sorted(requirements):
        versions = registry.get_compatible_registry_versions(
            versions, PackageSpec(owner=owner, name=name, requirements=requirement)
        )
    if not versions:
        raise RuntimeError(
            f"No version of {owner}/{name} satisfies all requirements "
            f"{sorted(requirements)} requested across the library tree"
        )

    best = registry.pick_best_registry_version(versions)
    pkgfile = registry.pick_compatible_pkg_file(best["files"])
    if not pkgfile:
        raise RuntimeError(f"No package file for {owner}/{name}@{best['name']}")
    return owner, name, best["name"], pkgfile["download_url"]


def _normalize_dependencies(dependencies: Any) -> list[dict]:
    """Normalize a library manifest's ``dependencies`` to a list of dicts.

    PIO's library.json accepts both the list-of-dicts form and the shorthand
    dict form (``{"owner/Name": "version_spec"}``); normalize the latter so
    callers see a uniform list.
    """
    if not dependencies:
        return []
    if isinstance(dependencies, dict):
        normalized = []
        for raw_name, spec in dependencies.items():
            if "/" in raw_name:
                owner, pkgname = raw_name.split("/", 1)
            else:
                owner, pkgname = None, raw_name
            entry = {"name": pkgname, "owner": owner}
            if isinstance(spec, dict):
                entry.update(spec)
            else:
                entry["version"] = spec
            normalized.append(entry)
        return normalized
    return [d for d in dependencies if isinstance(d, dict)]


@dataclass
class _LibNode:
    """A node in the library dependency graph being resolved as a batch."""

    key: str
    is_git: bool
    owner: str | None = None
    pkgname: str | None = None
    requirements: set[str] = field(default_factory=set)
    url: str | None = None
    ref: str | None = None
    edges: set[str] = field(default_factory=set)


def _url_or_none(value: Any) -> str | None:
    """Return ``value`` if it parses as a URL (scheme and host), else None."""
    if not value or not isinstance(value, str):
        return None
    try:
        parsed = urlsplit(value)
    except ValueError:
        return None
    return value if parsed.scheme and parsed.netloc else None


def _node_key(
    name: str | None, version: str | None, repository: str | None
) -> tuple[str, bool, tuple[str | None, str | None]]:
    """Return ``(key, is_git, locator)`` for a library or dependency spec.

    The key is derived from the *input* spec (the registry name as written, or
    the git URL path), not the resolved canonical name. So a package referenced
    inconsistently -- bare ``name`` vs ``owner/name``, or git vs registry -- maps
    to distinct keys and isn't deduplicated; ``convert_libraries`` warns about
    that after resolution rather than merging the nodes.

    PlatformIO's Library Manager also accepted a git URL in the *name*
    position (``add_library("https://github.com/x/y", None)``), including the
    ``git+`` VCS prefix and the ``CustomName=URL`` form; recognize those here
    so such specs resolve as git sources instead of failing a registry lookup.
    """
    if not repository and name and "://" in name:
        # Try the whole name first so a bare URL whose query contains ``=``
        # stays intact; fall back to the ``CustomName=URL`` form, where the
        # key derives from the URL path and the custom name is irrelevant.
        repository = _url_or_none(name) or _url_or_none(name.split("=", 1)[-1])
        if repository is None:
            # Anything with ``://`` was meant to be a URL; failing it fast
            # beats a confusing registry "package not found" error.
            raise RuntimeError(f"Invalid PIO library URL: {name}")
    if repository:
        split_result = urlsplit(repository.removeprefix("git+"))
        key = str(split_result.path).strip("/").removesuffix(".git")
        ref = split_result.fragment.strip() or None
        url = urlunsplit(split_result._replace(fragment=""))
        return key, True, (url, ref)
    if name and "/" in name:
        owner, pkgname = name.split("/", 1)
    else:
        owner, pkgname = None, name
    return name, False, (owner, pkgname)


def convert_libraries(
    libraries: list[Library], backend: LibraryBackend
) -> list[ConvertedLibrary]:
    """Resolve and convert a batch of PlatformIO libraries for ``backend``.

    Resolves the whole set together rather than each library independently: it
    walks the dependency graph collecting every version *requirement* per
    component name, then resolves each name once to a single version satisfying
    all of them. So a transitive dependency shared under
    different specs (e.g. ``esphome/libsodium``, pulled by both ``noise-c`` and
    ``esp_wireguard``) becomes one component instead of two clashing
    ``override_path`` entries -- order-independently, and without ever violating
    a stated constraint.

    The returned list holds the top-level components (those directly requested);
    transitive dependencies are converted too and wired into each component's
    generated manifest. ``backend.emit`` is called once per converted library to
    write its toolchain-specific build files.

    ``lib_ignore`` from ``esphome->platformio_options`` excludes libraries by
    short name (part after the ``/``), matched against both the top-level
    libraries and every dependency discovered during the graph walk.
    """
    nodes: dict[str, _LibNode] = {}

    lib_ignore = {
        name.split("/")[-1].lower()
        for name in CORE.platformio_options.get("lib_ignore", [])
    }

    # The generated build files inside the shared cache bake in the dependency
    # wiring, which lib_ignore changes; salt the cache path so configs with
    # different lib_ignore values don't fight over (and constantly rewrite) the
    # same converted component files.
    salt = (
        hashlib.sha256(",".join(sorted(lib_ignore)).encode()).hexdigest()[:8]
        if lib_ignore
        else ""
    )

    def is_ignored(name: str | None) -> bool:
        if not lib_ignore or name is None:
            return False
        return name.split("/")[-1].lower() in lib_ignore

    def add_spec(name: str | None, version: str | None, repository: str | None) -> str:
        key, is_git, locator = _node_key(name, version, repository)
        node = nodes.get(key) or _LibNode(key=key, is_git=is_git)
        nodes[key] = node
        if is_git:
            node.is_git = True
            node.url, node.ref = locator
        else:
            node.owner, node.pkgname = locator
            if version:
                node.requirements.add(version)
        return key

    top_level = [
        add_spec(library.name, library.version, library.repository)
        for library in libraries
        if not is_ignored(library.name)
    ]

    # Collect + resolve to a fixpoint: a node is (re)resolved whenever its
    # requirement set has grown since the last time, so every requirement in the
    # graph is accounted for before conversion.
    components: dict[str, ConvertedLibrary] = {}
    resolved_requirements: dict[str, frozenset[str]] = {}
    top_level_keys = set(top_level)
    worklist = deque(dict.fromkeys(top_level))
    while worklist:
        key = worklist.popleft()
        node = nodes[key]

        # A node is queued once per referring edge; skip the (uncached) registry
        # lookup + download + dependency walk unless its requirement set grew
        # since the last resolve. Requirements only ever grow, so this still
        # converges the fixpoint and terminates dependency cycles.
        requirements = frozenset(node.requirements)
        if resolved_requirements.get(key) == requirements:
            continue
        resolved_requirements[key] = requirements

        if node.is_git:
            component = ConvertedLibrary(key, "*", GitSource(node.url, node.ref))
        else:
            owner, name, version, url = _resolve_registry_version(
                node.owner, node.pkgname, node.requirements
            )
            component = ConvertedLibrary(
                _owner_pkgname_to_name(owner, name), version, URLSource(url)
            )
        component.download(salt=salt, namespace=backend.cache_key)

        library_json_path = component.path / "library.json"
        library_properties_path = component.path / "library.properties"
        has_json = library_json_path.is_file()
        has_properties = library_properties_path.is_file()
        if not has_json and not has_properties:
            # The shared cache can hold a broken copy (e.g. a clone or an
            # extraction interrupted by a killed process). Force one
            # re-download so a bad cache entry self-heals instead of failing
            # every build until the user runs a full clean.
            _LOGGER.warning(
                "Library %s at %s is missing library.json and library.properties; "
                "re-downloading",
                key,
                component.path,
            )
            component.download(force=True, salt=salt, namespace=backend.cache_key)
            has_json = library_json_path.is_file()
            has_properties = library_properties_path.is_file()
        if has_json:
            component.data = _parse_library_json(library_json_path)
        elif has_properties:
            component.data = _parse_library_properties(library_properties_path)
        else:
            raise RuntimeError(
                f"Invalid PIO library {key}: missing library.json and "
                f"library.properties in {component.path}"
            )

        try:
            check_library_data(component.data, backend.platform, backend.framework)
        except InvalidLibrary as e:
            # Skip an incompatible transitive dependency, but fail fast if a
            # top-level library the build explicitly requested is incompatible.
            if key in top_level_keys:
                raise RuntimeError(
                    f"Requested library {key} is not compatible with "
                    f"{backend.framework}: {e}"
                ) from e
            _LOGGER.debug("Skip incompatible dependency %s: %s", key, str(e))
            continue
        components[key] = component

        # Requirements changed (we got past the short-circuit above), so
        # (re)walk this component's dependencies.
        node.edges = set()
        for dependency in _normalize_dependencies(component.data.get("dependencies")):
            if "name" not in dependency or "version" not in dependency:
                continue
            try:
                check_library_data(dependency, backend.platform, backend.framework)
            except InvalidLibrary as e:
                _LOGGER.debug("Skip dependency %s: %s", dependency.get("name"), str(e))
                continue
            dep_name = _owner_pkgname_to_name(
                dependency.get("owner"), dependency.get("name")
            )
            if is_ignored(dep_name):
                _LOGGER.debug("Skip ignored dependency %s", dep_name)
                continue
            # The version field may actually be a URL (git/archive dependency).
            dep_version = dependency["version"]
            dep_url = _url_or_none(dep_version)
            if dep_url is not None:
                dep_version = None
            dep_key = add_spec(dep_name, dep_version, dep_url)
            node.edges.add(dep_key)
            worklist.append(dep_key)

    # A git source wins over any registry version requested for the same
    # component. That's intentional, but warn so a dropped registry pin isn't a
    # silent surprise.
    for node in nodes.values():
        if node.is_git and node.requirements:
            _LOGGER.warning(
                "Library %s is requested both from a git source (%s) and as "
                "registry version(s) %s; using the git source.",
                node.key,
                node.url,
                sorted(node.requirements),
            )

    # Two graph nodes that resolve to the same component name (e.g. a package
    # referenced both bare and as ``owner/name``) are not deduplicated and can
    # produce conflicting component definitions. Warn so it's not silent.
    canonical_keys: dict[str, str] = {}
    for node_key, component in components.items():
        canonical = component.get_sanitized_name()
        if canonical_keys.setdefault(canonical, node_key) != node_key:
            _LOGGER.warning(
                "Library %s is referenced under multiple names (%s and %s); these "
                "are not deduplicated. Reference it consistently as %s.",
                canonical,
                canonical_keys[canonical],
                node_key,
                canonical,
            )

    # Wire each component's dependencies to the single resolved instances, then
    # emit build files.
    for key, component in components.items():
        component.dependencies = [
            components[dep_key]
            for dep_key in sorted(nodes[key].edges)
            if dep_key in components
        ]
    for component in components.values():
        backend.emit(component)

    return [components[key] for key in top_level if key in components]
